#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#

"""
Functions to convert between .NET `DateTime` and similar instances and Python `pendulum` instances.
"""


import datetime as dt
import enum
import functools
from typing import Tuple, Union

import dateutil.tz as duz
import pendulum as pdt
import pendulum.tz as ptz

from orchid import base

# noinspection PyUnresolvedReferences,PyPackageRequirements
from System import (DateTime, DateTimeKind, DateTimeOffset, TimeSpan)


# Encapsulate the use of pendulum and DateTime.
UTC = pdt.UTC
NET_NAT = DateTime.MinValue
NAT = pdt.DateTime.min


class TimePointTimeZoneKind(enum.Enum):
    """Models the kind of time point.

    This class eases conversions to the .NET `DateTime` class by providing Python with similar capabilities as
    the .NET `Enum`. (See
    [DateTimeKind](https://docs.microsoft.com/en-us/dotnet/api/system.datetimekind?view=net-5.0) for details).
    """
    UTC = DateTimeKind.Utc  # Time zone is UTC
    LOCAL = DateTimeKind.Local  # Time zone is specified to be local
    UNSPECIFIED = DateTimeKind.Unspecified  # Time zone is unspecified


class NetDateTimeError(ValueError):
    """
    Raised when an error occurs accessing the .NET `TimeZoneInfo` of a .NET `DateTime` instance.
    """
    pass


class NetDateTimeLocalDateTimeKindError(NetDateTimeError):
    """
    Raised when the `DateTime.Kind` property of a .NET `DateTime` instance is `DateTimeKind.Local`.
    """
    def __init__(self, net_time_point: Union[DateTime, DateTimeOffset]):
        """
        Construct an instance from a .NET DateTime point in time.

        Args:
            net_time_point: A .NET DateTime representing a specific point in time.
        """
        super().__init__(self, '.NET DateTimeKind cannot be Local.', net_time_point.ToString("O"))


class NetDateTimeUnspecifiedDateTimeKindError(NetDateTimeError):
    """
    Raised when the `DateTimeKind` property of a .NET `DateTime` instance is not recognized.
    """
    ERROR_PREFACE = '.NET DateTimeKind is unexpectedly Unspecified.'

    ERROR_SUFFIX = """
    Although .NET DateTime.Kind should not be Unspecified, it may be
    safe to ignore this error by catching the exception.

    However, because it unexpected, **please** report the issue to
    Reveal Energy Services. 
    """

    def __init__(self, net_time_point: Union[DateTime, DateTimeOffset]):
        """
        Construct an instance from a .NET DateTime point in time.

        Args:
            net_time_point: A .NET DateTime representing a specific point in time.
        """
        super().__init__(self, NetDateTimeUnspecifiedDateTimeKindError.ERROR_PREFACE,
                         net_time_point.ToString("O"), NetDateTimeUnspecifiedDateTimeKindError.ERROR_SUFFIX)


class NetDateTimeNoTzInfoError(NetDateTimeError):
    """
    Raised when the `DateTimeKind` property of a .NET `DateTime` instance is
    `DateTimeKind.Unspecified`.
    """
    def __init__(self, time_point):
        """
        Construct an instance from a Python point in time.

        Args:
            time_point: A `pdt.DateTime` representing a specific point in time.
        """
        super().__init__(self, f'The Python time point must specify the time zone.', time_point.isoformat())


class NetDateTimeOffsetNonZeroOffsetError(NetDateTimeError):
    """
    Raised when the `Offset` property of a .NET `DateTimeOffset` is non-zero.
    """
    def __init__(self, net_date_time_offset):
        """
        Construct an instance from a .NET `DateTimeOffset`.

        Args:
            net_date_time_offset: A .NET `DateTimeOffset` representing a specific point in time.
        """
        super().__init__(self,
                         f'The `Offset` of the .NET `DateTimeOffset`, {net_date_time_offset.ToString("o")},'
                         ' cannot be non-zero.')


@functools.singledispatch
def as_date_time(net_time_point: object) -> pdt.DateTime:
    raise NotImplementedError


@as_date_time.register(type(None))
def _(net_time_point) -> pdt.DateTime:
    """
    Convert a .NET `DateTime` instance to a `pdt.DateTime` instance.

    Args:
        net_time_point: A point in time of type .NET `DateTime`.

    Returns:
        The `pdt.DateTime` equivalent to the `to_test`.

        If `net_time_point` is `DateTime.MaxValue`, returns `pdt.DateTime.max`. If `net_time_point` is
        `DateTime.MinValue`, returns `DATETIME_NAT`.
    """
    return NAT


@as_date_time.register
def _(net_time_point: DateTime) -> pdt.DateTime:
    """
    Convert a .NET `DateTime` instance to a `pdt.DateTime` instance.

    Args:
        net_time_point: A point in time of type .NET `DateTime`.

    Returns:
        The `pdt.DateTime` equivalent to the `to_test`.

        If `net_time_point` is `DateTime.MaxValue`, returns `pdt.DateTime.max`. If `net_time_point` is
        `DateTime.MinValue`, returns `DATETIME_NAT`.
    """
    if net_time_point == DateTime.MaxValue:
        return pdt.DateTime.max

    if net_time_point == DateTime.MinValue:
        return NAT

    if net_time_point.Kind == DateTimeKind.Utc:
        return _net_time_point_to_datetime(base.constantly(ptz.UTC), net_time_point)

    if net_time_point.Kind == DateTimeKind.Unspecified:
        raise NetDateTimeUnspecifiedDateTimeKindError(net_time_point)

    if net_time_point.Kind == DateTimeKind.Local:
        raise NetDateTimeLocalDateTimeKindError(net_time_point)

    raise ValueError(f'Unknown .NET DateTime.Kind, {net_time_point.Kind}.')


@as_date_time.register
def _(net_time_point: DateTimeOffset) -> pdt.DateTime:
    """
    Convert a .NET `DateTimeOffset` instance to a `pdt.DateTime` instance.

    Args:
        net_time_point: A point in time of type .NET `DateTimeOffset`.

    Returns:
        The `pdt.DateTime` equivalent to the `net_time_point`.
    """
    if net_time_point == DateTimeOffset.MaxValue:
        return pdt.DateTime.max

    if net_time_point == DateTimeOffset.MinValue:
        return NAT

    def net_date_time_offset_to_timezone(ntp):
        integral_offset = int(ntp.Offset.TotalSeconds)
        if integral_offset == 0:
            return ptz.UTC

        return ptz.timezone(integral_offset)

    return _net_time_point_to_datetime(net_date_time_offset_to_timezone, net_time_point)


def as_net_date_time(time_point: pdt.DateTime) -> DateTime:
    """
    Convert a `pdt.DateTime` instance to a .NET `DateTime` instance.

    Args:
        time_point: The `pdt.DateTime` instance to covert.

    Returns:
        The equivalent .NET `DateTime` instance.

        If `time_point` is `pdt.DateTime.max`, return `DateTime.MaxValue`. If `time_point` is
        `DATETIME_NAT`, return `DateTime.MinValue`.
    """
    if time_point == pdt.DateTime.max:
        return DateTime.MaxValue

    if time_point == NAT:
        return DateTime.MinValue

    if not time_point.tzinfo == ptz.UTC:
        raise NetDateTimeNoTzInfoError(time_point)

    carry_seconds, milliseconds = microseconds_to_milliseconds_with_carry(time_point.microsecond)
    result = DateTime(time_point.year, time_point.month, time_point.day,
                      time_point.hour, time_point.minute, time_point.second + carry_seconds,
                      milliseconds, DateTimeKind.Utc)
    return result


def as_net_date_time_offset(time_point: pdt.DateTime) -> DateTimeOffset:
    """
    Convert a `pdt.DateTime` instance to a .NET `DateTimeOffset` instance.

    Args:
        time_point: The `pdt.DateTime` instance to covert.

    Returns:
        The equivalent .NET `DateTimeOffset` instance.

        If `time_point` is `pdt.DateTime.max`, return `DateTime.MaxValue`. If `time_point` is
        `DATETIME_NAT`, return `DateTime.MinValue`.
    """
    if time_point == pdt.DateTime.max:
        return DateTimeOffset.MaxValue

    if time_point == NAT:
        return DateTimeOffset.MinValue

    date_time = as_net_date_time(time_point)
    result = DateTimeOffset(date_time)
    return result


def as_net_time_span(to_convert: pdt.Duration):
    """
    Convert a `pdt.Duration` instance to a .NET `TimeSpan`.

    Args:
        to_convert: The `pdt.Duration` instance to convert.

    Returns:
        The .NET `TimeSpan` equivalent to `to_convert`.
    """
    return TimeSpan(round(to_convert.total_seconds() * TimeSpan.TicksPerSecond))


def as_duration(to_convert: TimeSpan) -> pdt.Duration:
    """
    Convert a .NET `TimeSpan` to a python `pdt.Duration`

    Args:
        to_convert: The .NET `TimeSpan` to convert.

    Returns:
        The `pdt.Duration` equivalent to `to_convert`.

    """
    return pdt.duration(seconds=to_convert.TotalSeconds)


def as_time_delta(net_time_span: TimeSpan):
    """
    Convert a .NET `TimeSpan` to a Python `dt.timedelta`.

    Args:
        net_time_span: The .NET `TimeSpan` to convert.

    Returns:
        The equivalent dt.time_delta value.

    """
    return dt.timedelta(seconds=net_time_span.TotalSeconds)


def microseconds_to_milliseconds_with_carry(to_convert: int) -> Tuple[int, int]:
    """
    Convert microseconds to an integral number of milliseconds with a number of seconds to carry.

    Args:
        to_convert: The microseconds to convert.

    Returns:
        A tuple of the form, (number of seconds to "carry",  number of the integral milliseconds).
    """

    raw_milliseconds = round(to_convert / 1000)
    return divmod(raw_milliseconds, 1000)


def is_utc(time_point):
    return (time_point.tzinfo == pdt.UTC or
            time_point.tzinfo == dt.timezone.utc or
            time_point.tzinfo == duz.UTC)


def _net_time_point_to_datetime(time_zone_func, net_time_point):
    return pdt.datetime(net_time_point.Year, net_time_point.Month, net_time_point.Day,
                        net_time_point.Hour, net_time_point.Minute, net_time_point.Second,
                        net_time_point.Millisecond * 1000, tz=time_zone_func(net_time_point))



#  Copyright (c) 2017-2022 Reveal Energy Services, Inc 
#
#  Licensed under the Apache License, Version 2.0 (the "License"); 
#  you may not use this file except in compliance with the License. 
#  You may obtain a copy of the License at 
#
#      http://www.apache.org/licenses/LICENSE-2.0 
#
#  Unless required by applicable law or agreed to in writing, software 
#  distributed under the License is distributed on an "AS IS" BASIS, 
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
#  See the License for the specific language governing permissions and 
#  limitations under the License. 
#
# This file is part of Orchid and related technologies.
#

import uuid
from typing import Callable, Mapping, Union

import deal
import option
import toolz.curried as toolz

from orchid import (
    unit_system as units,
)

# noinspection PyUnresolvedReferences
from System import Guid

# These methods in this module are based on the StackOverflow post:
# https://stackoverflow.com/questions/36580931/python-property-factory-or-descriptor-class-for-wrapping-an-external-library
#
# Additionally, it resolves an issue I was experiencing with PyCharm: when I used `property` directly
# in the class definition, PyCharm reported "Property 'xyz' could not be read. I think it might have been
# than I needed to apply `curry` to the "getter method" I also defined in the class in order to pass the
# attribute name at definition time (because `self` was only available at run-time).


def get_dot_net_property_value(attribute_name, dom_object):
    """
    Return the value of the DOM property whose name corresponds to `attribute_name`.
    :param attribute_name: The Python `attribute_name`.
    :param dom_object: The DOM object whose property is sought.
    :return: The value of the DOM property.
    """
    @toolz.curry
    def python_name_to_words(python_name):
        return python_name.split('_')

    def capitalize_words(words):
        return toolz.map(str.capitalize, words)

    def words_to_dot_net_property_name(words):
        return ''.join(words)

    @toolz.curry
    def get_value_from_dom(dom, property_name):
        return getattr(dom, property_name)

    # The function, `thread_last`, from `toolz.curried`, "splices" threads a value (the first argument)
    # through each of the remaining functions as the *last* argument to each of these functions.
    result = toolz.thread_last(attribute_name,
                               python_name_to_words,
                               capitalize_words,
                               words_to_dot_net_property_name,
                               get_value_from_dom(dom_object))
    return result


def dom_property(attribute_name, docstring):
    """
    Return the property of the DOM corresponding to `attribute_name` with doc string.
    :param attribute_name: The name of the Python attribute.
    :param docstring: The doc string to be attached to the resulting property.
    :return: The Python property wrapping the value of the DOM property.
    """
    def getter(self):
        result = get_dot_net_property_value(attribute_name, self._adaptee)
        return result

    # Ensure no setter for the DOM properties
    return property(fget=getter, doc=docstring, fset=None)


def map_reduce_dom_property(attribute_name, docstring, mapper, reducer, initial):
    """
    Return reduced collection property of the DOM corresponding to `attribute_name` with doc string, `docstring`.

    Args:
        attribute_name: The name of the transformed attribute.
        docstring: The doc string to be attached to the resultant property.
        mapper: A callable that transforms every item of the .NET DOM property.
        reducer: A callable invoked to reduce the collection.
        initial: The initial value of the reduction.

    Returns:
        The Python property wrapping the mapped and reduced DOM (collection) property items.
    """
    def getter(self):
        result = toolz.pipe(get_dot_net_property_value(attribute_name, self._adaptee),
                            lambda container: container.Items,
                            toolz.map(mapper),
                            lambda items: toolz.reduce(reducer, items, initial))
        return result

    # Ensure no setter for the DOM properties
    return property(fget=getter, doc=docstring, fset=None)


def transformed_dom_property(attribute_name, docstring, transformer):
    """
    Return the transformed property of the DOM corresponding to `attribute_name`.
    :param attribute_name: The name of the Python attribute.
    :param docstring: The doc string to be attached to the resulting property.
    :param transformer: A callable to transform the value returned by the .NET DOM property.
    :return: The Python property wrapping the transformed value of the DOM property.
    """
    def getter(self):
        raw_result = get_dot_net_property_value(attribute_name, self._adaptee)
        result = transformer(raw_result)
        return result

    # Ensure no setter for the DOM properties
    return property(fget=getter, doc=docstring, fset=None)


def transformed_dom_property_iterator(attribute_name, docstring, transformer):
    """
    Return transformed collection property of the DOM corresponding to `attribute_name` with doc string, `docstring`.
    :param attribute_name: The name of the original attribute.
    :param docstring: The doc string to be attached to the resultant property.
    :param transformer: A callable invoked on each value in the list returned by the .NET DOM property.
    :return: The Python property wrapping a Python iterator mapping values from the DOM property (collection) items.
    """
    def getter(self):
        container = get_dot_net_property_value(attribute_name, self._adaptee)
        result = toolz.map(transformer, container.Items)
        return result

    # Ensure no setter for the DOM properties
    return property(fget=getter, doc=docstring, fset=None)


def as_object_id(guid: Guid):
    return uuid.UUID(str(guid))


class DotNetAdapter:
    @deal.pre(lambda _self, adaptee, _net_project_callable=None: adaptee is not None)
    def __init__(self, adaptee):
        """
        Construct an instance adapting the .NET `adaptee`.

        Args:
            adaptee: The .NET DOM object to adapt.
        """
        self._adaptee = adaptee

    @property
    def dom_object(self):
        """
        (PROTECTED) Determine the DOM object being adapted.

        This method is only intended to be used **INSIDE** the orchid package. It is **NOT** intended for
        external use.

        Returns:
            The .NET DOM object being adapted.
        """
        return self._adaptee


class IdentifiedDotNetAdapter(DotNetAdapter):
    def __init__(self, adaptee, net_project_callable: Callable = None):
        """
        Construct an instance adapting `adaptee` with access to the .NET `IProject` provided by `net_project_callable`.
        Args:
            adaptee: The .NET DOM object to adapt.
            net_project_callable: A callable returning the .NET `IProject` instance.
        """
        super().__init__(adaptee)
        self._net_project_callable = net_project_callable

    object_id = transformed_dom_property('object_id', 'The object ID of the adapted .NET DOM object.', as_object_id)

    @property
    def expect_project_units(self) -> Union[units.UsOilfield, units.Metric]:
        """
        (PROTECTED) Return the `UnitSystem` appropriate the .NET `IProject` of this instance.

        Although by naming convention, this property is "public," the author intends it to be "protected";
        that is, only called by classes derived from `IdentifiedDotNetAdapter` (and not necessarily all of those).

        Returns:
            A unit system - either `units.UsOilfield` or `units.Metric`.

        Raises:
            ValueError if the derived instances, for example, a `BaseTimeSeriesAdapter` is **not** associated with
            an `IProject`.
        """
        return self._maybe_project_units.expect('Expected associated project')

    @property
    def _maybe_project_units(self) -> option.Option[Union[units.UsOilfield, units.Metric]]:
        """
        Return the `option.Option[UnitSystem]` appropriate the .NET `IProject` of this instance.

        Although by naming convention, this property is "public," the author intends it to be "protected";
        that is, only called by classes derived from `IdentifiedDotNetAdapter` (and not necessarily all of those).

        Returns:
            An `option.Option()` instance.

            Some derived classes, such as 'NativeWellAdapter` and
            `NativeDataFrame`, contain a `Project` property that returns the `IProject` of the .NET class.
            For other derived classes, such as `BaseTimeSeriesAdapter`, the `IProject` cannot be determined. In the
            former situation, an associated `IProject`, this method returns an
            `option.Some[units.UsOilfield]` or an `option.Some[units.Metric]`. For the latter situation, no
            associated `IProject`, this method returns `option.NONE`. The methods of the
            [option.Option class](https://mat1g3r.github.io/option/option.html#module-option.option) allow
            derived classes to process these two situations with less likelihood of an error.
        """
        return (option.Some(units.as_unit_system(self._net_project_callable().ProjectUnits))
                if self._net_project_callable
                else option.NONE)


def dictionary_by_id(accumulator: Mapping[uuid.UUID, IdentifiedDotNetAdapter], mapped_object: IdentifiedDotNetAdapter):
    """
    Return a `Mapping` resulting from adding `mapped_object` to `accumulator`.

    Args:
        accumulator: The accumulated result.
        mapped_object: The `IdentifiedDotNetAdapter` to be added to `accumulator`.

    Returns:
        The updated `Mapping`.
    """
    return toolz.assoc(accumulator, mapped_object.object_id, mapped_object)



#  Copyright 2017-2022 Reveal Energy Services, Inc
#
#  Licensed under the Apache License, Version 2.0 (the "License"); 
#  you may not use this file except in compliance with the License. 
#  You may obtain a copy of the License at 
#
#      http://www.apache.org/licenses/LICENSE-2.0 
#
#  Unless required by applicable law or agreed to in writing, software 
#  distributed under the License is distributed on an "AS IS" BASIS, 
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
#  See the License for the specific language governing permissions and 
#  limitations under the License. 
#
# This file is part of Orchid and related technologies.
#

import json
from typing import Callable, Union
import uuid

import toolz.curried as toolz

from orchid import (
    dot_net_dom_access as dna,
    dot_net_disposable as dnd,
    net_stage_qc as nqc,
)


# noinspection PyUnresolvedReferences,PyPackageRequirements
from Orchid.FractureDiagnostics.Settings import IProjectUserData, Variant


class NativeProjectUserDataAdapter(dna.DotNetAdapter):
    """Adapts a .NET `IProjectUserData` instance to Python."""

    # TODO: Add code to `SdkAdapter` to allow handling `GetValue` and .NET `CorrectionStatus`
    # Python.NET has a [known issue](https://github.com/pythonnet/pythonnet/issues/1220) with its
    # handling of .NET Enum types. The Python.NET team has repaired this issue; however, the repair
    # is targeted for Python.NET 3. The team has specifically stated that they will *not* backport
    # this fix to the 2.5 version. Because .NET `CorrectionStatus` is a .NET `Enum` type, we
    # *cannot* construct a .NET `Variant` of type .NET `CorrectionStatus`.
    #
    # Our work-around for this issue is to implement this class in terms of the the JSON available
    # from the `IProjectUserData.ToJson`. This choice further requires hard-coding the logic used
    # by `StartStopTimeEditorViewModel` default `Variant` logic for QC notes and for start stop
    # confirmation.

    def __init__(self, adaptee: IProjectUserData):
        super().__init__(adaptee)

    def stage_qc_notes(self, stage_id: uuid.UUID) -> str:
        """
        Calculate the QC notes for the specified stage.

        Args:
            stage_id: The object ID of the stage of interest.

        Returns:
            The requested QC notes.
        """
        key_func = nqc.make_qc_notes_key
        transform_func = toolz.identity
        default_result = ''

        return self._extract_value_for_stage_id(stage_id, key_func, default_result, transform_func)

    def stage_start_stop_confirmation(self, stage_id: uuid.UUID) -> nqc.CorrectionStatus:
        """
        Calculate the start stop confirmation for the specified stage.

        Args:
            stage_id: The object ID of the stage of interest.

        Returns:
            The requested start stop confirmation.
        """
        key_func = nqc.make_start_stop_confirmation_key
        transform_func = nqc.CorrectionStatus
        default_result = nqc.CorrectionStatus.NEW

        return self._extract_value_for_stage_id(stage_id, key_func, default_result, transform_func)

    def _extract_value_for_stage_id(self,
                                    stage_id: uuid.UUID,
                                    key_func: Callable[[uuid.UUID], str],
                                    default_result: Union[str, nqc.CorrectionStatus],
                                    transform_func: Callable[[str], Union[str, nqc.CorrectionStatus]]):
        project_user_data_json = json.loads(self.dom_object.ToJson())
        confirmation_key = key_func(stage_id)
        # TODO: Replace hard-coded "copy" of logic
        # Hard-coded logic for QC notes default value from `StartStopTimeEditorViewModel`:
        # return .NET `CorrectionStatus.New` if either of stage ID or of start stop
        # confirmation is unavailable.
        result = default_result
        if confirmation_key in project_user_data_json:
            actual_value_type = toolz.get_in([confirmation_key, 'Type'], project_user_data_json)
            assert actual_value_type == 'System.String', (f'Expected, "System.String",'
                                                          f' but found "{actual_value_type}".')
            text_status = toolz.get_in([confirmation_key, 'Value'], project_user_data_json)
            result = transform_func(text_status)
        return result

    def set_stage_qc_notes(self, stage_id: uuid.UUID, to_notes: str) -> None:
        """
        Set the stage QC notes `to_notes` for the specified stage.
        Args:
            stage_id: The object ID that identifies the stage of interest.
            to_notes: The value to which to set the stage QC notes.
        """
        self._set_value(stage_id, nqc.make_qc_notes_key, to_notes, toolz.identity)

    def set_stage_start_stop_confirmation(self, stage_id: uuid.UUID,
                                          to_confirmation: nqc.CorrectionStatus) -> None:
        """
        Set the stage start stop confirmation `to_confirmation` for the specified stage.

        Args:
            stage_id: The object ID that identifies the stage of interest.
            to_confirmation: The value to which to set the stage start stop confirmation.
        """
        self._set_value(stage_id, nqc.make_start_stop_confirmation_key, to_confirmation, lambda v: v.value)

    def _set_value(self, stage_id, key_func, to_value, value_func):
        """
        Invoke `SetValue` on the mutable DOM object.

        Args:
            stage_id: The object ID identifying the stage whose value is to be set.
            key_func: A callable that generates the key identifying the value.
            to_value: The value to set.
            value_func: A function transforming the value to the appropriate .NET value.
        """
        with dnd.disposable(self.dom_object.ToMutable()) as mutable_pud:
            mutable_pud.SetValue(key_func(stage_id),
                                 Variant.Create.Overloads[str](value_func(to_value)))



#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#


import abc
from typing import Union

import toolz.curried as toolz
import option

from orchid import (
    dot_net_dom_access as dna,
    net_quantity as onq,
    unit_system as units,
)

# noinspection PyUnresolvedReferences,PyPackageRequirements
from Orchid.FractureDiagnostics import ISubsurfacePoint


class SubsurfacePoint(dna.IdentifiedDotNetAdapter):
    """An abstract base class for subsurface points."""

    def __init__(self, adaptee: ISubsurfacePoint, target_length_unit: Union[units.UsOilfield, units.Metric]):
        """
        Construct an instance adapting `adaptee` so that all lengths are expressed in `target_length_unit`.

        Args:
            adaptee: The .NET `ISubsurfacePoint` being adapted.
            target_length_unit: The target unit for all lengths.
        """
        super().__init__(adaptee)
        self._as_length_measurement_func = onq.as_measurement(target_length_unit)

    depth_origin = dna.dom_property('depth_datum',
                                    'The datum or origin for the z-coordinate of this point.')
    xy_origin = dna.dom_property('well_reference_frame_xy',
                                 'The reference frame or origin for the x-y coordinates of this point.')

    @property
    def x(self):
        """The x-coordinate of this point."""
        return self._as_length_measurement_func(option.maybe(self.dom_object.X))

    @property
    def y(self):
        """The y-coordinate of this point."""
        return self._as_length_measurement_func(option.maybe(self.dom_object.Y))

    @property
    def depth(self):
        """The depth of this point."""
        return self._as_length_measurement_func(option.maybe(self.dom_object.Depth))


@toolz.curry
def make_subsurface_point(target_unit: Union[units.UsOilfield, units.Metric],
                          net_subsurface_point: ISubsurfacePoint) -> SubsurfacePoint:
    return SubsurfacePoint(net_subsurface_point, target_unit)



#  Copyright (c) 2017-2022 Reveal Energy Services, Inc 
#
#  Licensed under the Apache License, Version 2.0 (the "License"); 
#  you may not use this file except in compliance with the License. 
#  You may obtain a copy of the License at 
#
#      http://www.apache.org/licenses/LICENSE-2.0 
#
#  Unless required by applicable law or agreed to in writing, software 
#  distributed under the License is distributed on an "AS IS" BASIS, 
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
#  See the License for the specific language governing permissions and 
#  limitations under the License. 
#
# This file is part of Orchid and related technologies.
#

from collections import namedtuple
from typing import Iterable, List, Tuple

import deal
import option
import toolz.curried as toolz

from orchid import (
    dot_net_dom_access as dna,
    searchable_project_objects as spo,
    native_data_frame_adapter as dfa,
    native_monitor_adapter as nma,
    native_time_series_adapter as tsa,
    native_project_user_data_adapter as uda,
    native_well_adapter as nwa,
    net_quantity as onq,
    unit_system as units,
)
from orchid.project_store import ProjectStore

# noinspection PyUnresolvedReferences,PyPackageRequirements
from Orchid.FractureDiagnostics import IWell, UnitSystem
# noinspection PyUnresolvedReferences,PyPackageRequirements
from Orchid.FractureDiagnostics.Settings import IProjectUserData
# noinspection PyUnresolvedReferences
import UnitsNet


ProjectBounds = namedtuple('ProjectBounds', [
    'min_x', 'max_x',
    'min_y', 'max_y',
    'min_depth', 'max_depth',
])
SurfacePoint = namedtuple('SurfacePoint', ['x', 'y'])


class Project(dna.IdentifiedDotNetAdapter):
    """Adapts a .NET `IProject` to a Pythonic interface."""

    @deal.pre(lambda self, project_loader: project_loader is not None)
    def __init__(self, project_loader: ProjectStore):
        """
        Construct an instance adapting he project available from net_project.

        :param project_loader: Loads an IProject to be adapted.
        """
        super().__init__(project_loader.native_project())
        self._project_loader = project_loader

    azimuth = dna.transformed_dom_property('azimuth', 'The azimuth of the project measured east of north.',
                                           toolz.compose(onq.as_measurement(units.Common.ANGLE),
                                                         option.maybe))
    name = dna.dom_property('name', 'The name of this project.')
    project_units = dna.transformed_dom_property('project_units', 'The project unit system.', units.as_unit_system)

    # _data_frames = dna.map_reduce_dom_property('data_frames', 'The project data frames.',
    #                                            dfa.NativeDataFrameAdapterIdentified, dna.dictionary_by_id, {})

    @property
    def fluid_density(self):
        """The fluid density of the project in project units."""
        return onq.as_measurement(self.project_units.DENSITY, option.maybe(self.dom_object.FluidDensity))

    def data_frames(self) -> spo.SearchableProjectObjects:
        """
        Return a `spo.SearchableProjectObjects` instance of all the data frames for this project.

        Returns:
            An `spo.SearchableProjectObjects` for all the data frames of this project.
        """
        return spo.SearchableProjectObjects(dfa.NativeDataFrameAdapterIdentified, self.dom_object.DataFrames.Items)

    def default_well_colors(self) -> List[Tuple[float, float, float]]:
        """
        Calculate the default well colors for this project.
        :return: A list of RGB tuples.
        """
        result = list(map(tuple, self._project_loader.native_project().PlottingSettings.GetDefaultWellColors()))
        return result

    def monitors(self) -> spo.SearchableProjectObjects:
        """
        Return a `spo.SearchableProjectObjects` instance of all the monitors for this project.

        Returns:
            An `spo.SearchableProjectObjects` for all the monitors of this project.
        """
        return spo.SearchableProjectObjects(nma.NativeMonitorAdapter, self.dom_object.Monitors.Items)

    def project_bounds(self) -> ProjectBounds:
        result = toolz.pipe(self.dom_object.GetProjectBounds(),
                            toolz.map(option.maybe),
                            toolz.map(onq.as_measurement(self.project_units.LENGTH)),
                            list,
                            lambda ls: ProjectBounds(*ls))
        return result

    def project_center(self) -> SurfacePoint:
        """
        Return the location of the project center on the surface measured in project units.
        """
        net_center = self.dom_object.GetProjectCenter()
        result = toolz.pipe(net_center,
                            toolz.map(option.maybe),
                            toolz.map(onq.as_measurement(self.project_units.LENGTH)),
                            list,
                            lambda ls: SurfacePoint(ls[0], ls[1]))
        return result

    def proppant_concentration_mass_unit(self):
        if self.project_units == units.UsOilfield:
            return units.UsOilfield.MASS
        elif self.project_units == units.Metric:
            return units.Metric.MASS
        else:
            raise ValueError(f'Unknown unit system: {self.project_units}')

    def slurry_rate_volume_unit(self):
        if self.project_units == units.UsOilfield:
            return units.UsOilfield.VOLUME
        elif self.project_units == units.Metric:
            return units.Metric.VOLUME
        else:
            raise ValueError(f'Unknown unit system: {self.project_units}')

    def time_series(self) -> spo.SearchableProjectObjects:
        """
        Return a `spo.SearchableProjectObjects` instance of all the time series for this project.

        Returns:
            An `spo.SearchableProjectObjects` for all the time series of this project.
        """
        return spo.SearchableProjectObjects(tsa.NativeTimeSeriesAdapter, self.dom_object.WellTimeSeriesList.Items)

    @property
    def user_data(self) -> uda.NativeProjectUserDataAdapter:
        return uda.NativeProjectUserDataAdapter(self.dom_object.ProjectUserData)

    def wells(self) -> spo.SearchableProjectObjects:
        """
        Return a `spo.SearchableProjectObjects` instance of all the wells for this project.

        Returns:
            An `spo.SearchableProjectObjects` for all the wells of this project.
        """
        return spo.SearchableProjectObjects(nwa.NativeWellAdapter, self.dom_object.Wells.Items)

    def wells_by_name(self, name) -> Iterable[IWell]:
        """
        Return all the wells in this project with the specified name.
        :param name: The name of the well(s) of interest.
        :return: A list of all the wells in this project.
        """
        return toolz.filter(lambda w: name == w.name, self.wells)



#  Copyright 2017-2022 Reveal Energy Services, Inc
#
#  Licensed under the Apache License, Version 2.0 (the "License"); 
#  you may not use this file except in compliance with the License. 
#  You may obtain a copy of the License at 
#
#      http://www.apache.org/licenses/LICENSE-2.0 
#
#  Unless required by applicable law or agreed to in writing, software 
#  distributed under the License is distributed on an "AS IS" BASIS, 
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
#  See the License for the specific language governing permissions and 
#  limitations under the License. 
#
# This file is part of Orchid and related technologies.
#


import option

from orchid import (
    base,
    dot_net_dom_access as dna,
    dom_project_object as dpo,
    measurement as om,
    net_date_time as ndt,
    net_quantity as onq,
)


class NativeStagePartAdapter(dpo.DomProjectObject):
    def __init__(self, adaptee):
        super().__init__(adaptee, base.constantly(adaptee.Project))

    display_name_with_well = dna.dom_property('display_name_with_well',
                                              'The display stage number including the well name')
    display_name_without_well = dna.dom_property('display_name_without_well',
                                                 'The display stage number excluding the well name')
    part_no = dna.dom_property('part_number', 'The part number for this stage part')
    start_time = dna.transformed_dom_property('start_time', 'The start time of this stage part',
                                              ndt.as_date_time)
    stop_time = dna.transformed_dom_property('stop_time', 'The stop time of this stage part',
                                             ndt.as_date_time)

    @property
    def isip(self) -> om.Quantity:
        """
        Return the instantaneous shut-in pressure of this stage part in project units.
        """
        return onq.as_measurement(self.expect_project_units.PRESSURE, option.maybe(self.dom_object.Isip))



#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#

import option

from orchid import dot_net_dom_access as dna


def transform_display_name(net_display_name):
    maybe_display_name = option.maybe(net_display_name)
    return maybe_display_name.expect('Unexpected value, `None`, for `display_name`.')


class DomProjectObject(dna.IdentifiedDotNetAdapter):
    name = dna.dom_property('name', 'The name of this data frame.')
    display_name = dna.transformed_dom_property('display_name', 'The display name of this data frame.',
                                                transform_display_name)



#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#

"""
This script illustrates using the Orchid* Python API, `pandas` and `matplotlib` to perform completion
 analysis.

(*Orchid is a mark of Reveal Energy Services, Inc)
"""

# Example: Using Pandas to Analyze Completion Parameters

# %% md

# 0.5 Import packages

# The only import needed for the Python API is `orchid` itself.

import orchid

# The remaining imports are standard python packages to support the analysis.

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
# The following import is included for its "side effects" of an improved color schemes and
# plot styles. (See the "Tip" in section 9.2 of "Python for Data Analysis" for details.)
import seaborn as sns


def compute_stage_treatment_aggregates(treatment_stage):
    """
    Compute the aggregate stage treatment data.

    Args:
        treatment_stage: The treatment stage whose aggregate data is sought.

    Returns:
        A `tuple` containing the treatment stage aggregate data.
    """

    # These calculations IGNORE all calculation warnings.
    pumped_fluid_volume = orchid.pumped_fluid_volume(treatment_stage,
                                                     treatment_stage.start_time,
                                                     treatment_stage.stop_time)[0].magnitude
    total_proppant_mass = orchid.total_proppant_mass(treatment_stage,
                                                     treatment_stage.start_time,
                                                     treatment_stage.stop_time)[0].magnitude
    median_treating_pressure = orchid.median_treating_pressure(treatment_stage,
                                                               treatment_stage.start_time,
                                                               treatment_stage.stop_time)[0].magnitude
    return pumped_fluid_volume, total_proppant_mass, median_treating_pressure


def show_proppant_loading_plot(data_frame, columns):
    """
    Build a proppant loading plot from `columns` of `data_frame`.

    Args:
        data_frame: Contains the data used to create the plot.
        columns: The columns containing the data to plot.
    """
    groups = data_frame.groupby(columns['well'])

    fig, ax = plt.subplots(len(groups), sharex='all', sharey='all', dpi=150)
    fig.suptitle(f'{columns["proppant_loading"]} by Stage')
    i = 0
    colors = ['tab:blue', 'tab:orange', 'tab:green', 'tab:red', 'tab:purple',
              'tab:brown', 'tab:pink', 'tab:gray', 'tab:olive', 'tab:cyan']
    for name, group in groups:
        ax[i].margins(0.05)  # Optional, just adds 5% padding to the autoscaling
        ax[i].plot(group[columns['stage']],
                   group[columns['proppant_loading']],
                   colors[i], marker='o', linestyle='', ms=6, label=name)
        ax[i].legend()
        ax[i].set_xlabel('Stage Number')
        i = i + 1
        # Hide x labels and tick labels for all but bottom plot.
    for a in ax:
        a.label_outer()
    plt.rcParams['figure.dpi'] = 150
    plt.show()


def to_unit(magnitude, src_unit, tgt_unit):
    """
    Helper function to convert units.

    Args:
        magnitude: The magnitude of the source.
        src_unit: The unit of measurement of the source.
        tgt_unit: The unit of measurement of the target.

    Returns:
        The magnitude of the source measurement measured in `tgt_unit`.
    """
    src_measurement = orchid.make_measurement(src_unit, magnitude)
    tgt_measurement = orchid.to_unit(tgt_unit, src_measurement)
    result = tgt_measurement.magnitude
    return result


def print_elided_data_frame(data_frame, title):
    print(title)
    print(data_frame.to_string(max_rows=10, max_cols=6))
    input('Press enter to continue...')
    print()


def analyze_completions():
    # 1.0 Load the .ifrac project

    # The following code simply captures the configured location of the Orchid training data. It is not needed
    # to use the Orchid Python API itself, but it is used in this example to load well - known data.
    orchid_training_data_path = orchid.training_data_path()

    bakken_project = orchid.load_project(str(orchid_training_data_path.joinpath(
        'frankNstein_Bakken_UTM13_FEET.ifrac')))

    # 3.0 Build a pandas data frame

    # Remember the project units
    bakken_units = bakken_project.project_units

    # Calculate the stage results
    stage_results = []
    for well in bakken_project.wells().all_objects():

        for stage in well.stages().all_objects():

            treatment_aggregates = compute_stage_treatment_aggregates(stage)
            # Skip stages with no aggregates. These stages most likely are from
            # an untreated monitor well.
            if not treatment_aggregates:
                continue

            stage_fluid, stage_proppant, median_stage_pressure = treatment_aggregates
            stage_results.append((bakken_project.name, well.name, stage.display_stage_number,
                                  stage.md_top(bakken_units.LENGTH).magnitude,
                                  stage.md_bottom(bakken_units.LENGTH).magnitude,
                                  stage_fluid, stage_proppant, median_stage_pressure))

    # Provide a way to manage DataFrame column names (which include units) using simpler,
    # semantic identifiers. Creating the column names helps me avoid "typos" involved in
    # getting the (project-specific) units correct, but is not necessary in a typical
    # interactive session.

    bakken_columns = {'project': 'Project',
                      'well': 'Well',
                      'stage': 'Stage',
                      'md_top': f'MD Top ({orchid.abbreviation(bakken_units.LENGTH)})',
                      'md_bottom': f'MD Bottom ({orchid.abbreviation(bakken_units.LENGTH)})',
                      'total_fluid':
                          f'Total Fluid ({orchid.abbreviation(bakken_project.slurry_rate_volume_unit())})',
                      'total_proppant':
                          f'Total Proppant ({orchid.abbreviation(bakken_project.proppant_concentration_mass_unit())})',
                      'median_treating':
                          f'Median Treating Pressure ({orchid.abbreviation(bakken_units.PRESSURE)})'}

    # Create the data frame
    bakken_summaries = pd.DataFrame(data=stage_results, columns=bakken_columns.values())
    print_elided_data_frame(bakken_summaries, 'Initial Bakken summaries data frame')

    # 3.1 Compute the stage length directly from the summary

    bakken_summaries[f'Stage Length ({orchid.abbreviation(bakken_units.LENGTH)})'] = \
        bakken_summaries.apply(lambda s: s[bakken_columns['md_bottom']] - s[bakken_columns['md_top']], axis=1)
    bakken_columns['stage_length'] = bakken_summaries.columns[-1]
    print_elided_data_frame(bakken_summaries, 'Bakken summaries with stage lengths')

    # 3.2 Now compute the proppant loading for each stage

    bakken_summaries[f'Proppant loading' \
                     f' ({orchid.abbreviation(bakken_project.proppant_concentration_mass_unit())}/'
                     f'{orchid.abbreviation(bakken_units.LENGTH)})'] = \
        bakken_summaries.apply(lambda s: s[bakken_columns['total_proppant']] / s[bakken_columns['stage_length']],
                               axis=1)
    bakken_columns['proppant_loading'] = bakken_summaries.columns[-1]
    print_elided_data_frame(bakken_summaries, 'Bakken summaries with proppant loading')

    # 4.0 Completion questions

    # 4.1 What is the median proppant intensity per well?

    bakken_summaries[[bakken_columns['well'],
                      bakken_columns['proppant_loading']]].groupby(bakken_columns['well']).median()

    groups = bakken_summaries.groupby(bakken_columns['well'])

    fig, ax = plt.subplots(dpi=150)
    ax.margins(0.05)  # Optional, just adds 5% padding to the autoscaling
    for name, group in groups:
        ax.plot(group[bakken_columns['stage']], group[bakken_columns['proppant_loading']],
                marker='o', linestyle='', ms=6, label=name)
    ax.legend()
    plt.show()

    # That's a little hard to interpret with the outliers

    # Plot
    groups = bakken_summaries[bakken_summaries[bakken_columns['stage']] > 5].groupby(bakken_columns['well'])

    fig, ax = plt.subplots(dpi=150)
    ax.margins(0.05)  # Optional, just adds 5% padding to the autoscaling
    for name, group in groups:
        ax.plot(group[bakken_columns['stage']], group[bakken_columns['proppant_loading']],
                marker='o', linestyle='', ms=6, label=name)
    ax.legend()
    ax.set_title(f'Proppant Loading by Stage')
    ax.set_xlabel('Stage Number')
    ax.set_ylabel(bakken_columns['proppant_loading'])
    plt.show()

    # That's a little busy. Let' s clean it up.

    show_proppant_loading_plot(bakken_summaries[bakken_summaries[bakken_columns['stage']] > 5], bakken_columns)

    # 4.2 What is the median treating pressure per well?

    bakken_summaries[[bakken_columns['well'],
                      bakken_columns['median_treating']]].groupby(bakken_columns['well']).median()

    groups = bakken_summaries.groupby(bakken_columns['well'])

    fig, ax = plt.subplots(3, sharex='all', sharey='all', dpi=150)
    fig.suptitle(f'{bakken_columns["median_treating"]} by Stage')
    i = 0
    colors = ['tab:blue', 'tab:orange', 'tab:green',
              'tab:purple', 'tab:yellow', 'tab:red', 'tab:black']
    for name, group in groups:
        ax[i].margins(0.05)
        ax[i].plot(group[bakken_columns['stage']],
                   group[bakken_columns['median_treating']],
                   colors[i], marker='o', linestyle='', ms=6, label=name)
        ax[i].legend()
        ax[i].set_xlabel('Stage Number')
        i = i + 1

        # Hide x labels and tick labels for all but bottom plot.
    for a in ax:
        a.label_outer()
    plt.show()

    # 5.0 Add another .ifrac project

    print('Wait patiently while I add another project...')
    print()

    montney_project = orchid.load_project(str(orchid_training_data_path.joinpath(
        'Project-frankNstein_Montney_UTM13_METERS.ifrac')))

    montney_units = montney_project.project_units

    # Calculate the stage results for the Montney project
    stage_results = []
    for well in montney_project.wells().all_objects():

        for stage in well.stages().all_objects():

            treatment_aggregates = compute_stage_treatment_aggregates(stage)
            # Skip stages with no aggregates. These stages most likely are from
            # an untreated monitor well.
            if not treatment_aggregates:
                continue

            stage_fluid, stage_proppant, median_stage_pressure = treatment_aggregates
            stage_results.append((montney_project.name, well.name, stage.display_stage_number,
                                  stage.md_top(montney_units.LENGTH).magnitude,
                                  stage.md_bottom(montney_units.LENGTH).magnitude,
                                  stage_fluid, stage_proppant, median_stage_pressure))

    # Provide a way to manage DataFrame column names (which include units) using simpler,
    # semantic identifiers.
    montney_columns = {'project': 'Project',
                       'well': 'Well',
                       'stage': 'Stage',
                       'md_top': f'MD Top ({orchid.abbreviation(montney_units.LENGTH)})',
                       'md_bottom': f'MD Bottom ({orchid.abbreviation(montney_units.LENGTH)})',
                       'total_fluid':
                           f'Total Fluid ('
                           f'{orchid.abbreviation(montney_project.slurry_rate_volume_unit())})',
                       'total_proppant':
                           f'Total Proppant ('
                           f'{orchid.abbreviation(montney_project.proppant_concentration_mass_unit())})',
                       'median_treating':
                           f'Median Treating Pressure ({orchid.abbreviation(montney_units.PRESSURE)})'}

    # Create the Montney data frame
    montney_summaries = pd.DataFrame(data=stage_results, columns=montney_columns.values())

    montney_summaries[f'Stage Length ({orchid.abbreviation(montney_units.LENGTH)})'] = \
        montney_summaries.apply(lambda s: s[montney_columns['md_bottom']] - s[montney_columns['md_top']], axis=1)
    montney_columns['stage_length'] = montney_summaries.columns[-1]

    montney_summaries[f'Proppant loading'
                      f' ({orchid.abbreviation(montney_project.proppant_concentration_mass_unit())}/'
                      f'{orchid.abbreviation(montney_units.LENGTH)})'] = \
        montney_summaries.apply(lambda s: s[montney_columns['total_proppant']] / s[montney_columns['stage_length']],
                                axis=1)
    montney_columns['proppant_loading'] = montney_summaries.columns[-1]

    print_elided_data_frame(montney_summaries, 'Montney summaries')

    # 5.1 Combine the data for the two projects.

    # The frankNstein_Bakken_UTM13_FEET project is in US Oilfield units, and the
    # "Project - frankNstein_Montney_UTM13_METERS" project is in metric units. I need to convert one to the other.

    # Calculate the stage results in US oilfield units for the Montney project
    stage_results = []
    for well in montney_project.wells().all_objects():

        for stage in well.stages().all_objects():

            treatment_aggregates = compute_stage_treatment_aggregates(stage)
            if not treatment_aggregates:
                continue

            # Convert stage aggregates to US oilfield units
            stage_fluid = to_unit(treatment_aggregates[0], montney_project.slurry_rate_volume_unit(),
                                  bakken_project.slurry_rate_volume_unit())
            stage_proppant = to_unit(treatment_aggregates[1], montney_project.proppant_concentration_mass_unit(),
                                     bakken_project.proppant_concentration_mass_unit())
            median_stage_pressure = to_unit(treatment_aggregates[2], montney_units.PRESSURE, bakken_units.PRESSURE)

            # Append results for this stage
            stage_results.append((montney_project.name, well.name,
                                  stage.display_stage_number,
                                  stage.md_top(bakken_units.LENGTH).magnitude,
                                  stage.md_bottom(bakken_units.LENGTH).magnitude,
                                  stage_fluid, stage_proppant, median_stage_pressure))

    # Create the converted Montney data frame
    montney_us_summaries = pd.DataFrame(data=stage_results,
                                        columns=list(bakken_columns.values())[:-2])

    montney_us_summaries[f'Stage Length ({orchid.abbreviation(bakken_units.LENGTH)})'] = \
        montney_us_summaries.apply(lambda s: s[bakken_columns['md_bottom']] - s[bakken_columns['md_top']], axis=1)

    montney_us_summaries[f'Proppant loading' 
                         f' ({orchid.abbreviation(bakken_project.proppant_concentration_mass_unit())}/'
                         f'{orchid.abbreviation(bakken_units.LENGTH)})'] = \
        montney_us_summaries.apply(lambda s: s[bakken_columns['total_proppant']] / s[bakken_columns['stage_length']],
                                   axis=1)

    print_elided_data_frame(montney_us_summaries, 'Montney summaries in US Oilfield units')

    combined = pd.concat([bakken_summaries[bakken_summaries[bakken_columns['stage']] > 5], montney_us_summaries])
    print_elided_data_frame(combined, 'Combined Bakken and Montney summaries')

    combined[[bakken_columns['well'],
              bakken_columns['proppant_loading']]].groupby(bakken_columns['well']).median()

    combined[[bakken_columns['well'],
              bakken_columns['proppant_loading']]] \
        .groupby(bakken_columns['well']) \
        .agg({bakken_columns['proppant_loading']: ['median', 'std']})

    show_proppant_loading_plot(combined, bakken_columns)


def main():
    print(__doc__)
    analyze_completions()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

#  Copyright (c) 2017-2022 Reveal Energy Services, Inc 
#
#  Licensed under the Apache License, Version 2.0 (the "License"); 
#  you may not use this file except in compliance with the License. 
#  You may obtain a copy of the License at 
#
#      http://www.apache.org/licenses/LICENSE-2.0 
#
#  Unless required by applicable law or agreed to in writing, software 
#  distributed under the License is distributed on an "AS IS" BASIS, 
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
#  See the License for the specific language governing permissions and 
#  limitations under the License. 
#
# This file is part of Orchid and related technologies.
#


from typing import Optional

import deal
import option

from orchid.project import Project
from orchid.project_store import ProjectStore

# To support doctests only
import shutil

import orchid


# TODO: change `ifrac_pathname` to be `str` or `pathlib.Path`
@deal.pre(lambda ifrac_pathname: len(ifrac_pathname.strip()) != 0)
@deal.pre(lambda ifrac_pathname: ifrac_pathname is not None)
def load_project(ifrac_pathname: str) -> Project:
    """
    Return the project for the specified `.ifrac` file.

    Args:
        ifrac_pathname: The path identifying the data file of the project of interest.

    Returns:
        The project of interest.

    Examples:
        >>> load_path = orchid.training_data_path().joinpath('frankNstein_Bakken_UTM13_FEET.ifrac')
        >>> loaded_project = orchid.load_project(str(load_path))
        >>> loaded_project.name
        'frankNstein_Bakken_UTM13_FEET'
    """
    loader = ProjectStore(ifrac_pathname.strip())
    result = Project(loader)
    return result


# TODO: change `ifrac_pathname` to be `str` or `pathlib.Path`
@deal.pre(lambda project, _: project is not None)
@deal.pre(lambda _, ifrac_pathname: ifrac_pathname is not None)
@deal.pre(lambda _, ifrac_pathname: len(ifrac_pathname) != 0)
@deal.pre(lambda _, ifrac_pathname: len(ifrac_pathname.strip()) != 0)
def save_project(project: Project, ifrac_pathname: str) -> None:
    """
    Return the project for the specified `.ifrac` file.

    Args:
        ifrac_pathname: The path identifying the data file of the project of interest.
        project: The project of interest.

    Examples:
        >>> # Test saving changed project
        >>> load_path = orchid.training_data_path().joinpath('frankNstein_Bakken_UTM13_FEET.ifrac')
        >>> loaded_project = orchid.load_project(str(load_path))
        >>> save_path = load_path.with_name(f'salvus{load_path.suffix}')
        >>> orchid.save_project(loaded_project, str(save_path))
        >>> save_path.exists()
        True
        >>> save_path.unlink()
    """

    store = ProjectStore(ifrac_pathname.strip())
    store.save_project(project)


# TODO: change `ifrac_pathname` to be `str` or `pathlib.Path`
@deal.pre(lambda _: _.project is not None)
@deal.pre(lambda _: _.project is not None)
@deal.pre(lambda _: _.source_pathname is not None)
@deal.pre(lambda _: len(_.source_pathname) != 0)
@deal.pre(lambda _: len(_.source_pathname.strip()) != 0)
@deal.pre(lambda _: (_.maybe_target_pathname is None or
                     (_.maybe_target_pathname is not None and len(_.maybe_target_pathname) != 0)))
@deal.pre(lambda _: (_.maybe_target_pathname is None or
                     (_.maybe_target_pathname is not None and len(_.maybe_target_pathname.strip()) != 0)))
def optimized_but_possibly_unsafe_save(project: Project, source_pathname: str,
                                       maybe_target_pathname: Optional[str] = None):
    """
    Saves `project`, optionally to `maybe_to_pathname` is an optimized, but possibly "unsafe" manner.

    If `maybe_to_pathname` is supplied and is not `None`, it must be a string representing a valid pathname. If a file
    with that pathname already exists, it will be overwritten (unless it is the same path as `source_pathname`)

    This method is unsafe because it only writes some data from `project`; the remainder of the data is simply
    (bulk) copied from the `.ifrac` file, `source_pathname`.

    This method assumes that `project` was originally loaded from `project_pathname` and was then changed in
    such a way that the "bulk" data **was not** changed. If this assumption is not true, the project saved in
    `to_pathname` will **not** contain all the changes to `project`.

    Specifically, this method **does not** save changes to data like:

    - Trajectories
    - Treatment curves
    - Monitor curves

    We believe that this method will generally finish more quickly than `save_project`; however, we cannot
    guarantee this behavior. We encourage the developer calling this method to perform her own performance tests
    and to understand if his use case meets the assumptions made by this method.

    Args:
        project: The project to be saved.
        source_pathname: The pathname of the `.ifrac` file from which `project` was loaded.
        maybe_target_pathname: The optional pathname of the `.ifrac` file in which to store `project`.

    Examples:
        >>> # Test optimized but possibly unsafe save of project
        >>> load_path = orchid.training_data_path().joinpath('Project_frankNstein_Permian_UTM13_FEET.ifrac')
        >>> loaded_project = orchid.load_project(str(load_path))
        >>> save_path = load_path.with_name(f'salva intuta{load_path.suffix}')
        >>> orchid.optimized_but_possibly_unsafe_save(loaded_project, str(load_path), str(save_path))
        >>> save_path.exists()
        True
        >>> save_path.unlink()
        >>> # Test optimized but possibly unsafe save of project in loaded location
        >>> source_path = orchid.training_data_path().joinpath('Project_frankNstein_Permian_UTM13_FEET.ifrac')
        >>> load_path = source_path.with_name(f'idem salvum filum{source_path.suffix}')
        >>> # ignore returned result for doctest
        >>> _to_path = shutil.copyfile(str(source_path), str(load_path))
        >>> loaded_project = orchid.load_project(str(load_path))
        >>> orchid.optimized_but_possibly_unsafe_save(loaded_project, str(load_path))
        >>> load_path.exists()
        True
        >>> load_path.unlink()

    """
    store = ProjectStore(source_pathname.strip())
    store.optimized_but_possibly_unsafe_save(project, option.maybe(maybe_target_pathname))



#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#
"""Wrapper function to help deal with IDisposable classes.

This implementation is based on code in the discussion of
[Python.NET issue 79](https://github.com/pythonnet/pythonnet/issues/79#issuecomment-187107566).
"""

from inspect import isclass, isfunction
from contextlib import contextmanager

__all__ = ['disposable']

"""Contextmanager wrapper around IDisposables."""


@contextmanager
def disposable(obj_or_class, *args, **kwargs):
    """
    Contextmanager wrapper around IDisposables.

    Example:

    >>> with disposable(dbConnection):
    >>>     blah()

    Args:
        obj_or_class: The .NET object or class with a `Dispose` method.
        *args: Positional arguments needed to manage the context managed by `__enter__/__exit__`.
        **kwargs: Keyword arguments needed to manage the context managed by `__exit__/__exit_-`.

    Returns:
        The instance of the object or class implementing `IDispose`.
    """

    if isclass(obj_or_class) or isfunction(obj_or_class):
        obj = obj_or_class(*args, **kwargs)
    else:
        obj = obj_or_class

    if hasattr(obj, '__enter__') and hasattr(obj, '__exit__'):
        # Already a contextmanager, skip
        return obj

    if not hasattr(obj, 'Dispose') or not callable(obj.Dispose):
        # Likely not IDisposable, currently isinstance doesn't work for
        # interfaces, so we have to check that in this explicit way
        return obj

    try:
        yield obj
    finally:
        obj.Dispose()



#  Copyright (c) 2017-2022 Reveal Energy Services, Inc 
#
#  Licensed under the Apache License, Version 2.0 (the "License"); 
#  you may not use this file except in compliance with the License. 
#  You may obtain a copy of the License at 
#
#      http://www.apache.org/licenses/LICENSE-2.0 
#
#  Unless required by applicable law or agreed to in writing, software 
#  distributed under the License is distributed on an "AS IS" BASIS, 
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
#  See the License for the specific language governing permissions and 
#  limitations under the License. 
#
# This file is part of Orchid and related technologies.
#


import os
import pathlib

import orchid.configuration
import orchid.script_adapter_context as sac

import toolz.curried as toolz

# noinspection PyPackageRequirements
import clr


def add_orchid_assemblies() -> None:
    """
    Add references to the Orchid assemblies needed by the Python API.

    Although not all modules in the `orchid` package need .NET types from all the available Orchid assemblies,
    I believe the additional cost of adding those references is far less than the cost of maintaining the
    copy-paste, boilerplate code that results without this common function.
    :return:
    """
    clr.AddReference('Orchid.Common')
    clr.AddReference('Orchid.FractureDiagnostics')
    clr.AddReference('Orchid.FractureDiagnostics.Factories')
    clr.AddReference('UnitsNet')
    clr.AddReference('System.Collections')
    return None


def app_settings_path() -> str:
    """
    Return the pathname of the `appSettings.json` file needed by the `SDKFacade` assembly.

    :return: The required pathname.
    """
    result = os.fspath(pathlib.Path(toolz.get_in(['orchid', 'root'],
                                                 orchid.configuration.python_api())).joinpath('appSettings.json'))
    return result


def prepare_imports() -> None:
    # This function call must occur *after*
    # - Importing clr
    # - Adding a reference to `Orchid.FractureDiagnostics.SDKFacade`
    # - Importing ScriptAdapter from `Orchid.FractureDiagnostics.SDKFacade`
    # - The call to `append_orchid_assemblies_directory_path`
    with sac.ScriptAdapterContext():
        orchid.dot_net.add_orchid_assemblies()



#  Copyright (c) 2017-2022 Reveal Energy Services, Inc
#
#  Licensed under the Apache License, Version 2.0 (the "License"); 
#  you may not use this file except in compliance with the License. 
#  You may obtain a copy of the License at 
#
#      http://www.apache.org/licenses/LICENSE-2.0 
#
#  Unless required by applicable law or agreed to in writing, software 
#  distributed under the License is distributed on an "AS IS" BASIS, 
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
#  See the License for the specific language governing permissions and 
#  limitations under the License. 
#
# This file is part of Orchid and related technologies.
#

import enum

import orchid.base
from orchid import (
    base_time_series_adapter as bca,
)

# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics.TimeSeries import IQuantityTimeSeries


# TODO: Replace with dataclass and constant class variables.
class TimeSeriesCurveTypes(enum.Enum):
    MONITOR_PRESSURE = 'Pressure'
    MONITOR_TEMPERATURE = 'Temperature'


class NativeTimeSeriesAdapter(bca.BaseTimeSeriesAdapter):
    def __init__(self, net_time_series: IQuantityTimeSeries):
        super().__init__(net_time_series, orchid.base.constantly(net_time_series.Well.Project))

    def quantity_name_unit_map(self, project_units):
        """
        Return a map (dictionary) between quantity names and units (from `unit_system`) of the data_points.

        This method plays the role of "Primitive Operation" in the *Template Method* design pattern. In this
        role, the "Template Method" defines an algorithm and delegates some steps of the algorithm to derived
        classes through invocation of "Primitive Operations".

        Args:
            project_units: The unit system of the project.
        """
        result = {
            TimeSeriesCurveTypes.MONITOR_PRESSURE.value: project_units.PRESSURE,
            TimeSeriesCurveTypes.MONITOR_TEMPERATURE.value: project_units.TEMPERATURE,
        }
        return result



#  Copyright (c) 2017-2022 Reveal Energy Services, Inc 
#
#  Licensed under the Apache License, Version 2.0 (the "License"); 
#  you may not use this file except in compliance with the License. 
#  You may obtain a copy of the License at 
#
#      http://www.apache.org/licenses/LICENSE-2.0 
#
#  Unless required by applicable law or agreed to in writing, software 
#  distributed under the License is distributed on an "AS IS" BASIS, 
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
#  See the License for the specific language governing permissions and 
#  limitations under the License. 
#
# This file is part of Orchid and related technologies.
#

"""This module contains common functions used to validate arguments."""


from orchid import (unit_system as units)


def arg_not_none(_, arg) -> bool:
    """
    Tests if the single argument is not None

    :param _: Ignored (typically mapped to `self` for bound methods)
    :param arg: The argument to be tested
    :return: True if arg is not None; otherwise, false.
    """
    return arg is not None


def arg_neither_empty_nor_all_whitespace(_, arg: str) -> bool:
    """
    Tests if the single argument is not None

    :param _: Ignored (typically mapped to `self` for bound methods)
    :param arg: The argument to be tested
    :return: True if arg is neither an empty string nor a string consisting only of whitespace.
    """
    return len(arg.strip()) > 0


def arg_is_acceptable_pressure_unit(_, target_unit):
    return (target_unit == units.UsOilfield.PRESSURE) or (target_unit == units.Metric.PRESSURE)


def is_unit_system_length(unit_to_test):
    return unit_to_test == units.UsOilfield.LENGTH or unit_to_test == units.Metric.LENGTH




#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#


from collections import namedtuple
import datetime as dt
from typing import Callable, Union

import deal
import option
import pendulum

from orchid import (
    project_store as loader,
    native_stage_adapter as nsa,
    net_date_time as net_dt,
    net_quantity as net_qty,
    unit_system as units,
)

#
# noinspection PyUnresolvedReferences,PyPackageRequirements
from Orchid.FractureDiagnostics import IStage
# noinspection PyUnresolvedReferences,PyPackageRequirements
from Orchid.FractureDiagnostics.Calculations import (ICalculationResult,
                                                     ITreatmentCalculations)
# noinspection PyUnresolvedReferences,PyPackageRequirements
from System import DateTime


CalculationResult = namedtuple('CalculationResult', ['measurement', 'warnings'])


def perform_calculation(native_calculation_func: Callable[[ITreatmentCalculations, IStage, DateTime, DateTime],
                                                          ICalculationResult],
                        stage: IStage,
                        start: DateTime,
                        stop: DateTime,
                        target_unit: Union[units.UsOilfield, units.Metric]):
    """
    Perform the specific native calculation function for stage from start through (and including) stop.

    Args:
        native_calculation_func: The specific native treatment calculation function.
        stage: The stage on which the calculation is being made.
        start: The (inclusive) start time of the calculation.
        stop: The (inclusive) stop time of the calculation.
        target_unit: The target unit of the measurement to be returned as the result.

    Returns:
        The calculation result (measurement and warnings) for the calculation.
    """
    native_treatment_calculations = loader.native_treatment_calculations()
    native_calculation_result = native_calculation_func(native_treatment_calculations, stage, start, stop)
    calculation_measurement = net_qty.as_measurement(target_unit,
                                                     option.maybe(native_calculation_result.Result))
    warnings = native_calculation_result.Warnings
    return CalculationResult(calculation_measurement, warnings)


@deal.pre(lambda _stage, start, _stop: net_dt.is_utc(start), message='Expected UTC for start time zone.')
@deal.pre(lambda _stage, _start, stop: net_dt.is_utc(stop), message='Expected UTC for stop time zone.')
def median_treating_pressure(stage: nsa.NativeStageAdapter,
                             start: Union[pendulum.DateTime, dt.datetime],
                             stop: Union[pendulum.DateTime, dt.datetime]):
    """
    Return the median treating pressure for stage from start to (and including) stop.

    Args:
        stage: The stage on which the calculation is being made.
        start: The (inclusive) start time of the calculation.
        stop: The (inclusive) stop time of the calculation.

    Returns:
        The median treating pressure result (measurement and warnings).
    """
    def median_treatment_pressure_calculation(calculations, for_stage, start_time, stop_time):
        calculation_result = calculations.GetMedianTreatmentPressure(for_stage.dom_object,
                                                                     net_dt.as_net_date_time(start_time),
                                                                     net_dt.as_net_date_time(stop_time))
        return calculation_result

    result = perform_calculation(median_treatment_pressure_calculation, stage,
                                 _datetime_to_pendulum(start), _datetime_to_pendulum(stop),
                                 stage.expect_project_units.PRESSURE)
    return result


@deal.pre(lambda _stage, start, _stop: net_dt.is_utc(start), message='Expected UTC for start time zone.')
@deal.pre(lambda _stage, _start, stop: net_dt.is_utc(stop), message='Expected UTC for stop time zone.')
def pumped_fluid_volume(stage: IStage,
                        start: Union[pendulum.DateTime, dt.datetime],
                        stop: Union[pendulum.DateTime, dt.datetime]):
    """
    Return the pumped (fluid) volume for stage from start to (and including) stop.

    Args:
        stage: The stage on which the calculation is being made.
        start: The (inclusive) start time of the calculation.
        stop: The (inclusive) stop time of the calculation.

    Returns:
        The pumped (fluid) volume result (measurement and warnings).
    """

    def pumped_fluid_volume_calculation(calculations, for_stage, start_time, stop_time):
        calculation_result = calculations.GetPumpedVolume(for_stage.dom_object,
                                                          net_dt.as_net_date_time(start_time),
                                                          net_dt.as_net_date_time(stop_time))
        return calculation_result

    result = perform_calculation(pumped_fluid_volume_calculation, stage,
                                 _datetime_to_pendulum(start), _datetime_to_pendulum(stop),
                                 stage.expect_project_units.VOLUME)
    return result


@deal.pre(lambda _stage, start, _stop: net_dt.is_utc(start), message='Expected UTC for start time zone.')
@deal.pre(lambda _stage, _start, stop: net_dt.is_utc(stop), message='Expected UTC for stop time zone.')
def total_proppant_mass(stage: IStage,
                        start: Union[pendulum.DateTime, dt.datetime],
                        stop: Union[pendulum.DateTime, dt.datetime]):
    """
    Return the pumped (fluid) volume for stage from start to (and including) stop.

    Args:
        stage: The stage on which the calculation is being made.
        start: The (inclusive) start time of the calculation.
        stop: The (inclusive) stop time of the calculation.

    Returns:
        The pumped (fluid) volume result (measurement and warnings).
    """
    def total_proppant_mass_calculation(calculations, for_stage, start_time, stop_time):
        calculation_result = calculations.GetTotalProppantMass(for_stage.dom_object,
                                                               net_dt.as_net_date_time(start_time),
                                                               net_dt.as_net_date_time(stop_time))
        return calculation_result

    result = perform_calculation(total_proppant_mass_calculation, stage,
                                 _datetime_to_pendulum(start), _datetime_to_pendulum(stop),
                                 stage.expect_project_units.MASS)
    return result


def _datetime_to_pendulum(source: dt.datetime) -> pendulum.DateTime:
    return pendulum.instance(source).set(tz=pendulum.UTC)



#  Copyright (c) 2017-2022 Reveal Energy Services, Inc 
#
#  Licensed under the Apache License, Version 2.0 (the "License"); 
#  you may not use this file except in compliance with the License. 
#  You may obtain a copy of the License at 
#
#      http://www.apache.org/licenses/LICENSE-2.0 
#
#  Unless required by applicable law or agreed to in writing, software 
#  distributed under the License is distributed on an "AS IS" BASIS, 
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
#  See the License for the specific language governing permissions and 
#  limitations under the License. 
#
# This file is part of Orchid and related technologies.
#

"""This module contains functions for converting between instances of the (Python) `Measurement` class and
instances of .NET classes like `UnitsNet.Quantity` and `DateTime`."""

from functools import singledispatch
import math
from numbers import Real
import operator
from typing import Optional, Union

import option
import toolz.curried as toolz

from orchid import (
    measurement as om,
    physical_quantity as opq,
    unit_system as units,
)

# noinspection PyUnresolvedReferences
from Optional import Option
# noinspection PyUnresolvedReferences
import System
# noinspection PyUnresolvedReferences
import UnitsNet


# Convenience functions


def to_net_quantity_value(magnitude):
    """
    Create a `UnitsNet` `QuantityValue` instance.

    The `UnitsNet` package *does not* accept floating point values to create most measurements; instead, these
    creation functions expect an argument of type `QuantityValue`. In .NET, these values are created using an
    implicit conversion. This implicit version on the Python side fails.

    This function allows us to define that conversion in one place.

    Args:
        magnitude: The magnitude of the `UnitNet` `Quantity` (measurement) to be created.

    Returns:
        The `magnitude` wrapped in a `UnitsNet.QuantityValue` instance.

    """
    return UnitsNet.QuantityValue.op_Implicit(magnitude)


def net_length_from(magnitude: float, net_unit: UnitsNet.Length.Units) -> UnitsNet.Quantity:
    """
    Create a `UnitsNet` length measurement from `magnitude` and `net_unit`.

    Args:
        magnitude: The magnitude of the length measurement
        net_unit: The `UnitsNet` unit of the length measurement

    Returns:
        The `UnitsNet` `Quantity` whose `Value` is `magnitude` and whose `Unit` is `net_unit`.
    """
    return UnitsNet.Length.From(to_net_quantity_value(magnitude), net_unit)


def net_pressure_from(magnitude: float, net_unit: UnitsNet.Pressure.Units) -> UnitsNet.Quantity:
    """
    Create a `UnitsNet` pressure measurement from `magnitude` and `net_unit`.

    Args:
        magnitude: The magnitude of the pressure measurement
        net_unit: The `UnitsNet` unit of the pressure measurement

    Returns:
        The `UnitsNet` `Quantity` whose `Value` is `magnitude` and whose `Unit` is `net_unit`.
    """
    return UnitsNet.Pressure.From(to_net_quantity_value(magnitude), net_unit)


# The following code creates conversion functions programmatically by:
# - Creating a map from variable name to string identifying how to create the `UnitsNet` `Quantity`
# - Transforming that map by:
#   - Transforming the map keys by prepending 'net_'
#   - Transforming the map values to a function creating the value. (See the documentation of
#     `operator.attrgetter` function for details.)
# - Adding the key and value of this map to the `globals()` dict (module attributes)
net_creator_attributes = {
    'angle_from_deg': 'Angle.FromDegrees',
    'duration_from_min': 'Duration.FromMinutes',
    'density_from_lbs_per_cu_ft': 'Density.FromPoundsPerCubicFoot',
    'density_from_kg_per_cu_m': 'Density.FromKilogramsPerCubicMeter',
    'energy_from_ft_lbs': 'Energy.FromFootPounds',
    'energy_from_J': 'Energy.FromJoules',
    'force_from_lbf': 'Force.FromPoundsForce',
    'force_from_N': 'Force.FromNewtons',
    'length_from_ft': 'Length.FromFeet',
    'length_from_m': 'Length.FromMeters',
    'mass_from_lbs': 'Mass.FromPounds',
    'mass_from_kg': 'Mass.FromKilograms',
    'power_from_hp': 'Power.FromMechanicalHorsepower',
    'power_from_W': 'Power.FromWatts',
    'pressure_from_psi': 'Pressure.FromPoundsForcePerSquareInch',
    'pressure_from_kPa': 'Pressure.FromKilopascals',
    'pressure_from_bars': 'Pressure.FromBars',
    'mass_concentration_from_lbs_per_gal': 'MassConcentration.FromPoundsPerUSGallon',
    'mass_concentration_from_kg_per_cu_m': 'MassConcentration.FromKilogramsPerCubicMeter',
    'volume_flow_from_oil_bbl_per_min': 'VolumeFlow.FromOilBarrelsPerMinute',
    'volume_flow_from_cu_m_per_min': 'VolumeFlow.FromCubicMetersPerMinute',
    'temperature_from_deg_F': 'Temperature.FromDegreesFahrenheit',
    'temperature_from_deg_C': 'Temperature.FromDegreesCelsius',
    'volume_from_oil_bbl': 'Volume.FromOilBarrels',
    'volume_from_cu_m': 'Volume.FromCubicMeters',
}
net_creator_funcs = toolz.pipe(
    net_creator_attributes,
    toolz.keymap(lambda k: f'net_{k}'),
    toolz.valmap(lambda v: toolz.compose(operator.attrgetter(v)(UnitsNet),
                                         to_net_quantity_value)),
)
for variable_name, variable_value in net_creator_funcs.items():
    globals()[variable_name] = variable_value


class EqualsComparisonDetails:
    def __init__(self, tolerance: Real = 1e-4,
                 net_comparison_type: UnitsNet.ComparisonType = UnitsNet.ComparisonType.Relative):
        """
        Construct an instance that uses `tolerance` and `comparison_type` to determine equality.

        This class exists because the `tolerance` and `comparison_type` are closely coupled; that is, one
        cannot correctly interpret the use of `tolerance` without a reference to the `comparison_type`.

        Args:
            tolerance: The maximum difference tolerated between two instances in determining equality.
            net_comparison_type: The type of comparison: `UnitsNet.ComparisonType.Relative` or
            `UnitsNet.ComparisonType.Absolute`.
        """
        self._tolerance = tolerance
        self._comparison_type = net_comparison_type

    @property
    def tolerance(self) -> Real:
        """
        Return the tolerance to be use in determining equality.

        Although this property is public, it is intended only to be read by the `equal_net_comparison`
        function.
        """
        return self._tolerance

    @property
    def comparison_type(self) -> UnitsNet.ComparisonType:
        """
        Return the comparison type to be use in determining equality.

        Although this property is public, it is intended only to be read by the `equal_net_comparison`
        function.
        """
        return self._comparison_type


#
# Although Pint supports the unit `cu_ft`, we have chosen to use the synonym, `ft ** 3` (which is
# printed as 'ft\u00b3` (that is, 'ft' followed by a Unicode superscript 3)). According to a
# citation on [Wikipedia article](https://en.wikipedia.org/wiki/Cubic_foot), this "is the IEEE
# symbol for the cubic foot." Our general rule: we accept the Pint unit `cu_ft` as **input**,
# but, on various conversion, produce the Pint unit `ft**3`.
#


@singledispatch
@toolz.curry
def as_measurement(unknown, _maybe_net_quantity: option.Option[UnitsNet.IQuantity]) -> om.Quantity:
    """
    Convert an optional .NET `IQuantity` to a `pint` `Quantity` instance.

    This function is registered as the type-handler for the `object` type. In our situation, arriving here
    indicates an error by an implementer and so raises an error.

    Args:
        unknown: A parameter whose type is not expected.
        _maybe_net_quantity: The optional .NET `IQuantity` instance to convert. (Unused.)
    """
    raise TypeError(f'First argument, {unknown}, has type {type(unknown)}, unexpected by `as_measurement`.')


# noinspection PyUnresolvedReferences
@as_measurement.register(units.Common)
@toolz.curry
def as_measurement_in_common_unit(target_unit, maybe_net_quantity: UnitsNet.IQuantity) -> om.Quantity:
    """
    Convert an optional .NET `IQuantity` to a `pint` `Quantity` instance in a common unit.

    Args:
        target_unit: The unit (from the units.Common) for the converted `Quantity` instance.
        maybe_net_quantity: The optional .NET `IQuantity` instance to convert.

    Returns:
        The equivalent `Quantity` instance in the target unit.
    """
    return maybe_net_quantity.map_or(_as_measurement_in_unit(target_unit),
                                     om.Quantity(float('NaN'), target_unit.value.unit))


@as_measurement.register(units.Metric)
@as_measurement.register(units.UsOilfield)
@toolz.curry
def as_measurement_in_specified_unit(target_unit,
                                     maybe_net_quantity: option.Option[UnitsNet.IQuantity]) -> om.Quantity:
    """
    Convert an optional .NET `IQuantity` to a `pint` `Quantity` instance.

    Args:
        target_unit: The unit for the converted `Quantity` instance.
        maybe_net_quantity: The optional .NET `IQuantity` instance to convert.

    Returns:
        The equivalent `Quantity` instance in the target unit.
    """

    result = maybe_net_quantity.map_or(_as_measurement_in_unit(target_unit),
                                       om.Quantity(float('NaN'), target_unit.value.unit))
    return result


@toolz.curry
def _as_measurement_in_unit(target_unit: Union[units.Metric, units.UsOilfield],
                            net_quantity: UnitsNet.IQuantity) -> om.Quantity:
    """
    Convert an `IQuantity` to a `pint` `Quantity` in a specified compatible unit.

    Args:
        target_unit: The target unit for the converted `Quantity` instance.
        net_quantity: The .NET `IQuantity` instance to convert.

    Returns:
        The equivalent `Quantity` instance in the specified unit.
    """
    target_magnitude = net_quantity.As(_UNIT_NET_UNITS[target_unit])
    result = om.Quantity(target_magnitude, target_unit.value.unit)
    return result


# noinspection PyUnresolvedReferences
_physical_quantity_to_net_physical_quantity = {
    opq.PhysicalQuantity.ANGLE: UnitsNet.Angle,
    opq.PhysicalQuantity.DURATION: UnitsNet.Duration,
    opq.PhysicalQuantity.DENSITY: UnitsNet.Density,
    opq.PhysicalQuantity.ENERGY: UnitsNet.Energy,
    opq.PhysicalQuantity.FORCE: UnitsNet.Force,
    opq.PhysicalQuantity.LENGTH: UnitsNet.Length,
    opq.PhysicalQuantity.MASS: UnitsNet.Mass,
    opq.PhysicalQuantity.POWER: UnitsNet.Power,
    opq.PhysicalQuantity.PRESSURE: UnitsNet.Pressure,
    opq.PhysicalQuantity.PROPPANT_CONCENTRATION: UnitsNet.MassConcentration,
    opq.PhysicalQuantity.TEMPERATURE: UnitsNet.Temperature,
    opq.PhysicalQuantity.SLURRY_RATE: UnitsNet.VolumeFlow,
    opq.PhysicalQuantity.VOLUME: UnitsNet.Volume,
}


@toolz.curry
def _python_measurement_option(target_unit: Union[units.Metric, units.UsOilfield],
                               optional_net_quantity: Option[UnitsNet.IQuantity]):
    @toolz.curry
    def net_option_as_pythonic(net_option):
        # If we are **not** an instance of `Option<Pressure>`
        if not hasattr(net_option, 'HasValue'):
            # Simply return the value (which may be `null` / `None`)
            return net_option

        # If we **are** an instance of `Option<Pressure>` yet have no value
        if not net_option.HasValue:
            return None

        # The variable, `net_quantity`, will contain a zero value in the `UnitsNet` unit corresponding to the Python
        # physical quantity of the `target_unit`. However, because of the preceding `if` statement, I expect this
        # value to **not** be returned. (It is actually an error if it is returned; unfortunately, it is an
        # undetectable error because I cannot distinguish the zero value from a zero value actually wrapped in
        # the `Option<T>.Some` expression.
        net_physical_quantity = _physical_quantity_to_net_physical_quantity[target_unit.value.physical_quantity]
        result = net_option.ValueOr.Overloads[net_physical_quantity](net_physical_quantity.Zero)
        return result

    return toolz.pipe(
        optional_net_quantity,
        net_option_as_pythonic,
        option.maybe,
    )


def as_measurement_from_option(target_unit: Union[units.Common, units.Metric, units.UsOilfield],
                               optional_net_quantity: Option[UnitsNet.IQuantity]):
    maybe_python_measurement = _python_measurement_option(target_unit, optional_net_quantity)
    return as_measurement(target_unit, maybe_python_measurement)


@singledispatch
@toolz.curry
def as_net_quantity(unknown, _measurement: om.Quantity) -> Optional[UnitsNet.IQuantity]:
    """
    Convert a .NET UnitsNet.IQuantity to a `pint` `Quantity` instance.

    This function is registered as the type-handler for the `object` type. In our situation, arriving here
    indicates an error by an implementer and so raises an error.

    Args:
        unknown: A parameter whose type is not expected.
        _measurement: The `Quantity` instance to convert.

    Returns:
        The equivalent `UnitsNet.IQuantity` instance.
    """
    raise TypeError(f'First argument, {unknown}, has type {type(unknown)}, unexpected by `as_net_quantity`.')


# noinspection PyUnresolvedReferences
_PINT_UNIT_CREATE_NET_UNITS = {
    om.registry.deg: UnitsNet.Angle.FromDegrees,
    om.registry.min: UnitsNet.Duration.FromMinutes,
    om.registry.ft_lb: UnitsNet.Energy.FromFootPounds,
    om.registry.J: UnitsNet.Energy.FromJoules,
    om.registry.lbf: UnitsNet.Force.FromPoundsForce,
    om.registry.N: UnitsNet.Force.FromNewtons,
    om.registry.ft: UnitsNet.Length.FromFeet,
    om.registry.m: UnitsNet.Length.FromMeters,
    om.registry.lb: UnitsNet.Mass.FromPounds,
    om.registry.kg: UnitsNet.Mass.FromKilograms,
    om.registry.hp: UnitsNet.Power.FromMechanicalHorsepower,
    om.registry.W: UnitsNet.Power.FromWatts,
    om.registry.psi: UnitsNet.Pressure.FromPoundsForcePerSquareInch,
    om.registry.kPa: UnitsNet.Pressure.FromKilopascals,
    om.registry.oil_bbl / om.registry.min: net_volume_flow_from_oil_bbl_per_min,
    ((om.registry.m ** 3) / om.registry.min): net_volume_flow_from_cu_m_per_min,
    om.registry.degF: UnitsNet.Temperature.FromDegreesFahrenheit,
    om.registry.degC: UnitsNet.Temperature.FromDegreesCelsius,
    om.registry.oil_bbl: UnitsNet.Volume.FromOilBarrels,
    (om.registry.m ** 3): UnitsNet.Volume.FromCubicMeters,
}


def _us_oilfield_slurry_rate(qv):
    return UnitsNet.Density.FromPoundsPerCubicFoot(qv)


# noinspection PyUnresolvedReferences
_PHYSICAL_QUANTITY_PINT_UNIT_NET_UNITS = {
    opq.PhysicalQuantity.DENSITY: {
        om.registry.lb / om.registry.cu_ft: _us_oilfield_slurry_rate,
        om.registry.lb / om.registry.ft ** 3: _us_oilfield_slurry_rate,
        om.registry.kg / (om.registry.m ** 3): UnitsNet.Density.FromKilogramsPerCubicMeter,
    },
    opq.PhysicalQuantity.PROPPANT_CONCENTRATION: {
        om.registry.lb / om.registry.gal: net_mass_concentration_from_lbs_per_gal,
        om.registry.kg / (om.registry.m ** 3): net_mass_concentration_from_kg_per_cu_m,
    },
}


# noinspection PyUnresolvedReferences
@as_net_quantity.register(opq.PhysicalQuantity)
@toolz.curry
def as_net_quantity_using_physical_quantity(physical_quantity,
                                            measurement: om.Quantity) -> Optional[UnitsNet.IQuantity]:
    """
    Convert a `Quantity` instance to a .NET `UnitsNet.IQuantity` instance.

    Args:
        physical_quantity: The `PhysicalQuantity`. Although we try to determine a unique mapping between units
        in `pint` and .NET `UnitsNet` units, we cannot perform a unique mapping for density and proppant
        concentration measured in the metric system (the units of both these physical quantities are
        "kg/m**3").
        measurement: The `Quantity` instance to convert.

    Returns:
        The equivalent `UnitsNet.IQuantity` instance.
    """
    if math.isnan(measurement.magnitude):
        return None

    quantity = UnitsNet.QuantityValue.op_Implicit(measurement.magnitude)
    if physical_quantity == opq.PhysicalQuantity.DENSITY:
        return toolz.get_in([physical_quantity, measurement.units], _PHYSICAL_QUANTITY_PINT_UNIT_NET_UNITS)(quantity)

    if physical_quantity == opq.PhysicalQuantity.PROPPANT_CONCENTRATION:
        return toolz.get_in([physical_quantity, measurement.units],
                            _PHYSICAL_QUANTITY_PINT_UNIT_NET_UNITS)(measurement.magnitude)

    if physical_quantity == opq.PhysicalQuantity.SLURRY_RATE:
        return toolz.get(measurement.units, _PINT_UNIT_CREATE_NET_UNITS)(measurement.magnitude)

    return toolz.get(measurement.units, _PINT_UNIT_CREATE_NET_UNITS)(quantity)


# noinspection PyUnresolvedReferences
@as_net_quantity.register(units.Common)
@toolz.curry
def as_net_quantity_using_common_units(to_common_unit, measurement: om.Quantity) -> Optional[UnitsNet.IQuantity]:
    """
    Convert a `Quantity` instance to a .NET `UnitsNet.IQuantity` instance corresponding `to_unit`.

    Args:
        to_common_unit: The target unit of measurement.
        measurement: The `Quantity` instance to convert.

    Returns:
        The equivalent `UnitsNet.IQuantity` instance.
    """
    # units.Common support no conversion so simply call another implementation.
    return as_net_quantity(to_common_unit.value.physical_quantity, measurement)


# noinspection PyUnresolvedReferences
@as_net_quantity.register(units.Metric)
@as_net_quantity.register(units.UsOilfield)
@toolz.curry
def as_net_quantity_in_specified_unit(specified_unit, measurement: om.Quantity) -> Optional[UnitsNet.IQuantity]:
    """
    Convert a `pint` `Quantity` to a .NET UnitsNet.IQuantity instance in a specified, but compatible unit.

    Args:
        specified_unit: The unit for the converted `Quantity` instance.
        measurement: The `Quantity` instance to convert.

    Returns:
        The equivalent `Quantity` instance in the specified unit.
    """
    target_measurement = measurement.to(specified_unit.value.unit)
    return as_net_quantity(specified_unit.value.physical_quantity, target_measurement)


def equal_net_quantities(left_quantity: UnitsNet.IQuantity, right_quantity: UnitsNet.IQuantity,
                         comparison_details: EqualsComparisonDetails = EqualsComparisonDetails()):
    """
    Compares two UnitsNet.IQuantity instances for equality

    Python.NET transforms == (perhaps indirectly) into a call to Equals. Unfortunately, comparing
    two measurements that have been transformed may have floating point differences. Specifically,
    UnitsNet marks the `Equals` method as `Obsolete` with the following message:
    > "It is not safe to compare equality due to using System.Double as the internal representation.
    > It is very easy to get slightly different values due to floating point operations. Instead use
    > Equals(Length, double, ComparisonType) to provide the max allowed absolute or relative error."

    Consequently, to determine if two `UnitsNet.IQuantity` instances are equal, I use the
    `Equals(Length, double, ComparisonType)` method applied to each instance.

    Args:
        left_quantity: The `IQuantity` instance on the "left-hand-side" of the (implicit) == operator.
        right_quantity: The `IQuantity` instance on the "right-hand-side" of the (implicit) == operator.
        comparison_details: The details of how to compare the two `UnitsNet.IQuantity` instances.

    Returns:

    """
    return left_quantity.Equals(right_quantity, comparison_details.tolerance, comparison_details.comparison_type)


def net_decimal_to_float(net_decimal: System.Decimal) -> float:
    """
    Convert a .NET Decimal value to a Python float.

    Python.NET currently leaves .NET values of type `Decimal` unconverted. For example, UnitsNet models units
    of the physical quantity, power, as values of type .NET 'QuantityValue` whose `Value` property returns a
    value of .NET `Decimal` type. This function assists in converting those values to Python values of type
    `float`.

    Args:
        net_decimal: The .NET `Decimal` value to convert.

    Returns:
        A value of type `float` that is "equivalent" to the .NET `Decimal` value. Note that this conversion is
        "lossy" because .NET `Decimal` values are exact, but `float` values are not.
    """
    return System.Decimal.ToDouble(net_decimal)


_UNIT_NET_UNITS = {
    units.Common.ANGLE: UnitsNet.Units.AngleUnit.Degree,
    units.Common.DURATION: UnitsNet.Units.DurationUnit.Minute,
    units.UsOilfield.DENSITY: UnitsNet.Units.DensityUnit.PoundPerCubicFoot,
    units.Metric.DENSITY: UnitsNet.Units.DensityUnit.KilogramPerCubicMeter,
    units.UsOilfield.ENERGY: UnitsNet.Units.EnergyUnit.FootPound,
    units.Metric.ENERGY: UnitsNet.Units.EnergyUnit.Joule,
    units.UsOilfield.FORCE: UnitsNet.Units.ForceUnit.PoundForce,
    units.Metric.FORCE: UnitsNet.Units.ForceUnit.Newton,
    units.UsOilfield.LENGTH: UnitsNet.Units.LengthUnit.Foot,
    units.Metric.LENGTH: UnitsNet.Units.LengthUnit.Meter,
    units.UsOilfield.MASS: UnitsNet.Units.MassUnit.Pound,
    units.Metric.MASS: UnitsNet.Units.MassUnit.Kilogram,
    units.UsOilfield.POWER: UnitsNet.Units.PowerUnit.MechanicalHorsepower,
    units.Metric.POWER: UnitsNet.Units.PowerUnit.Watt,
    units.UsOilfield.PRESSURE: UnitsNet.Units.PressureUnit.PoundForcePerSquareInch,
    units.Metric.PRESSURE: UnitsNet.Units.PressureUnit.Kilopascal,
    units.UsOilfield.PROPPANT_CONCENTRATION: UnitsNet.Units.MassConcentrationUnit.PoundPerUSGallon,
    units.Metric.PROPPANT_CONCENTRATION: UnitsNet.Units.MassConcentrationUnit.KilogramPerCubicMeter,
    units.UsOilfield.SLURRY_RATE: UnitsNet.Units.VolumeFlowUnit.OilBarrelPerMinute,
    units.Metric.SLURRY_RATE: UnitsNet.Units.VolumeFlowUnit.CubicMeterPerMinute,
    units.UsOilfield.TEMPERATURE: UnitsNet.Units.TemperatureUnit.DegreeFahrenheit,
    units.Metric.TEMPERATURE: UnitsNet.Units.TemperatureUnit.DegreeCelsius,
    units.UsOilfield.VOLUME: UnitsNet.Units.VolumeUnit.OilBarrel,
    units.Metric.VOLUME: UnitsNet.Units.VolumeUnit.CubicMeter,
}


@toolz.curry
def _convert_net_quantity_to_different_unit(target_unit: units.UnitSystem,
                                            net_quantity: UnitsNet.IQuantity) -> UnitsNet.IQuantity:
    """
    Convert one .NET `UnitsNet.IQuantity` to another .NET `UnitsNet.IQuantity` in a different unit `target_unit`
    Args:
        net_quantity: The `UnitsNet.IQuantity` instance to convert.
        target_unit: The unit to which to convert `maybe_net_quantity`.

    Returns:
        The .NET `UnitsNet.IQuantity` converted to `target_unit`.
    """
    result = net_quantity.ToUnit(_UNIT_NET_UNITS[target_unit])
    return result


def _net_decimal_to_float(net_decimal: System.Decimal) -> float:
    """
    Convert a .NET Decimal value to a Python float.

    Python.NET currently leaves .NET values of type `Decimal` unconverted. For example, UnitsNet models units
    of the physical quantity, power, as values of type .NET 'QuantityValue` whose `Value` property returns a
    value of .NET `Decimal` type. This function assists in converting those values to Python values of type
    `float`.

    Args:
        net_decimal: The .NET `Decimal` value to convert.

    Returns:
        A value of type `float` that is "equivalent" to the .NET `Decimal` value. Note that this conversion is
        "lossy" because .NET `Decimal` values are exact, but `float` values are not.
    """
    return System.Decimal.ToDouble(net_decimal)


_PHYSICAL_QUANTITY_NET_UNIT_PINT_UNITS = {
    opq.PhysicalQuantity.DENSITY: {
        UnitsNet.Units.DensityUnit.PoundPerCubicFoot: om.registry.lb / om.registry.ft ** 3,
        UnitsNet.Units.DensityUnit.KilogramPerCubicMeter: om.registry.kg / (om.registry.m ** 3),
    },
    opq.PhysicalQuantity.ENERGY: {
        UnitsNet.Units.EnergyUnit.FootPound: om.registry.ft_lb,
        UnitsNet.Units.EnergyUnit.Joule: om.registry.J,
    },
    opq.PhysicalQuantity.FORCE: {
        UnitsNet.Units.ForceUnit.PoundForce: om.registry.lbf,
        UnitsNet.Units.ForceUnit.Newton: om.registry.N,
    },
    opq.PhysicalQuantity.LENGTH: {
        UnitsNet.Units.LengthUnit.Foot: om.registry.ft,
        UnitsNet.Units.LengthUnit.Meter: om.registry.m,
    },
    opq.PhysicalQuantity.MASS: {
        UnitsNet.Units.MassUnit.Pound: om.registry.lb,
        UnitsNet.Units.MassUnit.Kilogram: om.registry.kg,
    },
    opq.PhysicalQuantity.POWER: {
        UnitsNet.Units.PowerUnit.MechanicalHorsepower: om.registry.hp,
        UnitsNet.Units.PowerUnit.Watt: om.registry.W,
    },
    opq.PhysicalQuantity.PRESSURE: {
        UnitsNet.Units.PressureUnit.PoundForcePerSquareInch: om.registry.psi,
        UnitsNet.Units.PressureUnit.Kilopascal: om.registry.kPa,
    },
    opq.PhysicalQuantity.PROPPANT_CONCENTRATION: {
        UnitsNet.Units.MassConcentrationUnit.PoundPerUSGallon: om.registry.lb / om.registry.gallon,
        UnitsNet.Units.MassConcentrationUnit.KilogramPerCubicMeter: om.registry.kg / om.registry.m ** 3,
    },
    opq.PhysicalQuantity.SLURRY_RATE: {
        UnitsNet.Units.VolumeFlowUnit.OilBarrelPerMinute: om.registry.oil_bbl / om.registry.min,
        UnitsNet.Units.VolumeFlowUnit.CubicMeterPerMinute: om.registry.m ** 3 / om.registry.min,
    },
    opq.PhysicalQuantity.TEMPERATURE: {
        UnitsNet.Units.TemperatureUnit.DegreeFahrenheit: om.registry.degF,
        UnitsNet.Units.TemperatureUnit.DegreeCelsius: om.registry.degC,
    },
    opq.PhysicalQuantity.VOLUME: {
        UnitsNet.Units.VolumeUnit.OilBarrel: om.registry.oil_bbl,
        UnitsNet.Units.VolumeUnit.CubicMeter: om.registry.m ** 3,
    },
}


def _to_pint_unit(physical_quantity: opq.PhysicalQuantity, net_unit: UnitsNet.Units) -> om.Unit:
    """
    Convert `net_unit`, a unit of measure for `physical_quantity`, to a `pint` unit.

    Args:
        physical_quantity: The physical quantity measured by `net_unit`.
        net_unit: The .NET UnitsNet.Unit to be converted.

    Returns:
        The `pint` Unit corresponding to `net_unit`.
    """
    result = toolz.get_in([physical_quantity, net_unit], _PHYSICAL_QUANTITY_NET_UNIT_PINT_UNITS)
    if result is not None:
        return result
    elif physical_quantity == opq.PhysicalQuantity.ANGLE:
        return om.registry.deg
    elif physical_quantity == opq.PhysicalQuantity.DURATION:
        return om.registry.min



#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#

"""Demonstrates using the high-level Orchid Python API to query and change the stage QC status."""

import argparse
import dataclasses as dc
import logging
import pathlib
from typing import Iterable

import orchid
from orchid import (
    project as op,
    net_stage_qc as nqc,
)


@dc.dataclass
class WellStagePair:
    well_name: str
    stage_no: int


@dc.dataclass
class StageQCInfo:
    start_stop_confirmation: nqc.CorrectionStatus
    stage_qc_notes: str


@dc.dataclass
class StageQCResult:
    well_stage_pair: WellStagePair
    stage_qc_info: StageQCInfo


def read_stage_qc(project: op.Project, sample_stages: Iterable[WellStagePair]) -> Iterable[StageQCResult]:
    """
    Read the stage QC result for each item of `sample_stages` of `project`.

    Args:
        project: The project of interest
        sample_stages: The iterable returning each `WellStagePair` whose stage QC results are sought.

    Returns:
        An iterable containing the `StageQCResult` for each `WellStagePair`
    """
    def read_stage_cq_result(project: op.Project, well_name_stage_pair: WellStagePair) -> StageQCResult:
        candidate_wells = list(project.wells().find_by_name(well_name_stage_pair.well_name))
        assert len(candidate_wells) == 1, (f'Expected 1 well with name, {well_name_stage_pair.well_name}.'
                                           f' Found {len(candidate_wells)}.')
        well = candidate_wells[0]
        stage = well.stages().find_by_display_stage_number(well_name_stage_pair.stage_no)

        project_user_data = project.user_data
        qc_notes = project_user_data.stage_qc_notes(stage.object_id)
        start_stop_confirmation = project_user_data.stage_start_stop_confirmation(stage.object_id)

        return StageQCResult(well_name_stage_pair, StageQCInfo(start_stop_confirmation=start_stop_confirmation,
                                                               stage_qc_notes=qc_notes))

    result = [read_stage_cq_result(project, wsp) for wsp in sample_stages]
    return result


def log_stage_qc_results(stage_qc_results: Iterable[StageQCResult], prefix: str) -> None:
    """
    Logs each item in `stage_qc_results`.

    Args:
        stage_qc_results: An iterable of `StageQCResult` instances.
        prefix: Text to write before writing the results.
    """
    logging.info(prefix)
    for stage_qc_result in stage_qc_results:
        logging.info(stage_qc_result)


def change_stage_qc(project: op.Project, to_stage_qc_results: Iterable[StageQCResult]) -> None:
    """
    Changes the stage QC results of `project` to `to_stage_qc_results`.

    Args:
        project: The project whose stage QC results are to be changed.
        to_stage_qc_results: The revised stage QC results.
    """
    for to_stage_qc_result in to_stage_qc_results:
        candidate_wells = list(project.wells().find_by_name(to_stage_qc_result.well_stage_pair.well_name))
        assert len(candidate_wells) == 1, (f'Expected 1 well with name,'
                                           f' {to_stage_qc_result.well_stage_pair.well_name}.'
                                           f' Found {len(candidate_wells)}.')
        well = candidate_wells[0]
        stage = well.stages().find_by_display_stage_number(to_stage_qc_result.well_stage_pair.stage_no)

        project_user_data = project.user_data
        project_user_data.set_stage_qc_notes(stage.object_id, to_stage_qc_result.stage_qc_info.stage_qc_notes)
        project_user_data.set_stage_start_stop_confirmation(stage.object_id,
                                                            to_stage_qc_result.stage_qc_info.start_stop_confirmation)


def main(cli_args):
    """
    Change stage start/stop times of stage 1 of well, `Demo_1H`.

    Args:
        cli_args: The command line arguments from `argparse.ArgumentParser`.
    """
    logging.basicConfig(level=logging.INFO)

    # Read Orchid project
    project = orchid.load_project(cli_args.input_project)

    # If no stages have undergone the stage QC process (like this project), reading the stage QC information produces
    # "uninteresting" results.
    sample_stages = [WellStagePair('Demo_1H', 9),
                     WellStagePair('Demo_2H', 7),
                     WellStagePair('Demo_4H', 23)]
    uninteresting_stage_qc_results = read_stage_qc(project, sample_stages)
    if cli_args.verbosity >= 1:
        log_stage_qc_results(uninteresting_stage_qc_results,
                             f'Reading results from input file: {cli_args.input_project}')

    if not cli_args.read_only:
        # Change the stage QC information of these same stages to be "interesting"
        to_stage_qc_info = [StageQCInfo(nqc.CorrectionStatus.UNCONFIRMED, stage_qc_notes='Strange'),
                            StageQCInfo(nqc.CorrectionStatus.CONFIRMED, stage_qc_notes='Good stage'),
                            StageQCInfo(nqc.CorrectionStatus.NEW, stage_qc_notes='')]
        to_stage_qc_results = [StageQCResult(wsp, info) for wsp, info in zip(sample_stages, to_stage_qc_info)]
        change_stage_qc(project, to_stage_qc_results)

        # Reading these same stages now produces "interesting" results
        interesting_stage_qc_results = read_stage_qc(project, sample_stages)
        if cli_args.verbosity >= 1:
            log_stage_qc_results(interesting_stage_qc_results, f'Reading results after change:')

        # Save project changes to specified .ifrac file
        orchid.optimized_but_possibly_unsafe_save(project, cli_args.input_project, cli_args.output_project)
        if cli_args.verbosity >= 1:
            logging.info(f'Wrote changes to "{cli_args.output_project}"')


def make_project_path_name(project_dir_name, project_file_name):
    """
    Make a path name to a project.

    Args:
        project_dir_name: The directory name of the project.
        project_file_name: The file name of the project.

    Returns:
        The path name to the .ifrac file for this project.
    """
    return str(project_dir_name.joinpath(project_file_name))


def make_target_file_name_from_source(source_file_name):
    """
    Make a file name for the changed project file name from the original project file name.

    Args:
        source_file_name: The file name of the project originally read.

    Returns:
        The project file name with a `.994` suffix inserted before the `.ifrac` suffix.
    """
    return ''.join([source_file_name.stem, '.994', source_file_name.suffix])


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Demonstrate reading and updating stage QC information.")
    parser.add_argument('-r', '--read-only', action='store_true', default=False,
                        help='Only read the stage QC results.')
    parser.add_argument('-v', '--verbosity', type=int, choices=[0, 1, 2], default=0,
                        help='Increase output verbosity. (Default: 0; that is, least output.)')

    parser.add_argument('input_project', help=f'Path name of project to read.')

    # Although input file must be 'frankNstein_Bakken_UTM13_FEET.v11.ifrac', I use the original version as the
    # default for the correct output file name.
    default_file_name_to_read = pathlib.Path('frankNstein_Bakken_UTM13_FEET.ifrac')
    default_project_path_name_to_read = make_project_path_name(orchid.training_data_path(),
                                                               default_file_name_to_read)
    default_file_name_to_write = make_target_file_name_from_source(default_file_name_to_read)
    default_project_path_name_to_write = make_project_path_name(orchid.training_data_path(),
                                                                default_file_name_to_write)
    parser.add_argument('-o', '--output_project', default=default_project_path_name_to_write,
                        help=f'Filename of project to write. (Default: {default_project_path_name_to_write}')

    args = parser.parse_args()
    main(args)




#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#


from enum import Enum

# noinspection PyUnresolvedReferences
import UnitsNet


class PhysicalQuantity(Enum):
    """The enumeration of physical quantities available via the Orchid Python API."""

    ANGLE = 'angle'
    DENSITY = 'density'
    DURATION = 'duration'
    ENERGY = 'energy'
    FORCE = 'force'
    LENGTH = 'length'
    MASS = 'mass'
    POWER = 'power'
    PRESSURE = 'pressure'
    PROPPANT_CONCENTRATION = 'proppant concentration'
    SLURRY_RATE = 'slurry rate'
    TEMPERATURE = 'temperature'
    VOLUME = 'volume'



#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#


from typing import Iterable, Optional


from orchid import (
    native_stage_part_adapter as spa,
    searchable_project_objects as spo,
)


class SearchableStageParts(spo.SearchableProjectObjects):
    def find_by_part_number(self, to_find: int) -> Optional[spa.NativeStagePartAdapter]:
        """
        Find a stage part with the specified part number.

        Args:
            to_find: The part number sought.

        Returns:
            The stage part with the specified part number.

            If no such stage part is found, returns `None`. If multiple stage parts with the specified part number are
            found, raises `spo.SearchableProjectMultipleMatchError`.
        """
        candidates = list(self.find(lambda s: s.part_no == to_find))
        if len(candidates) == 0:
            return None
        elif len(candidates) == 1:
            return candidates[0]
        else:
            raise spo.SearchableProjectMultipleMatchError(to_find)

    def find_by_display_name_with_well(self, to_find: str) -> Iterable[spa.NativeStagePartAdapter]:
        """
        Find a stage part with the specified display name with well.

        Args:
            to_find: The display name with well sought.

        Returns:
            The matching stage part(s).

            If no such stage part is in this collection, returns an empty iterator.
        """
        return self.find(lambda s: s.display_name_with_well == to_find)

    def find_by_display_name_without_well(self, to_find: str) -> Iterable[spa.NativeStagePartAdapter]:
        """
        Find a stage part without the specified display name with well.

        Args:
            to_find: The display name without well sought.

        Returns:
            The matching stage part(s).

            If no such stage part is in this collection, returns an empty iterator.
        """
        return self.find(lambda s: s.display_name_without_well == to_find)



#  Copyright (c) 2017-2022 Reveal Energy Services, Inc 
#
#  Licensed under the Apache License, Version 2.0 (the "License"); 
#  you may not use this file except in compliance with the License. 
#  You may obtain a copy of the License at 
#
#      http://www.apache.org/licenses/LICENSE-2.0 
#
#  Unless required by applicable law or agreed to in writing, software 
#  distributed under the License is distributed on an "AS IS" BASIS, 
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
#  See the License for the specific language governing permissions and 
#  limitations under the License. 
#
# This file is part of Orchid and related technologies.
#

import functools
import pathlib
from typing import Union

import deal
import option
import toolz.curried as toolz

from orchid import (
    dot_net,
    dot_net_disposable as dnd,
    script_adapter_context as sac,
    validation,
)

# noinspection PyUnresolvedReferences
from System import InvalidOperationException
# noinspection PyUnresolvedReferences
from System.IO import (FileStream, FileMode, FileAccess, FileShare)
# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics.SDKFacade import (
    PythonTimesSeriesArraysDto,
    ScriptAdapter,
)
# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics.TimeSeries import IQuantityTimeSeries

# To support doctests only
import json
import shutil
import zipfile

import orchid


class OrchidError(Exception):
    pass


# Ensure that a pathname is a string. Useful especially for converting `pathlib.Path` instances.
pathname_to_str = toolz.compose(str, toolz.identity)


def as_python_time_series_arrays(native_time_series: IQuantityTimeSeries):
    """
    Calculate the Python time series arrays equivalent to the `native_time_series` samples.
    Args:
        native_time_series: The native time series whose samples are sought.

    Returns:
        A `PythonTimeSeriesArraysDto` containing two arrays:
        - Sample magnitudes
        - Unix time stamps in seconds
    """
    with sac.ScriptAdapterContext():
        result = ScriptAdapter.AsPythonTimeSeriesArrays(native_time_series)
    return result


@functools.lru_cache()
def native_treatment_calculations():
    """
    Returns a .NET ITreatmentCalculations instance to be adapted.

    Returns:
            An `ITreatmentCalculations` implementation.
    """
    with sac.ScriptAdapterContext():
        result = ScriptAdapter.CreateTreatmentCalculations()
    return result


class ProjectStore:
    """Provides an .NET IProject to be adapted."""

    # TODO: consider changing `project_pathname` to be `pathlib.Path`
    @deal.pre(validation.arg_not_none)
    @deal.pre(validation.arg_neither_empty_nor_all_whitespace)
    def __init__(self, project_pathname: str):
        """
        Construct an instance that loads project data from project_pathname

        Args:
            project_pathname: Identifies the data file for the project of interest.
        """
        self._project_pathname = pathlib.Path(project_pathname)
        self._native_project = None
        self._in_context = False

    def native_project(self):
        """
        Return the native (.NET) Orchid project.

        Returns:
            The loaded `IProject`.
        """
        if self._native_project is None:
            self.load_project()
        return self._native_project

    def load_project(self):
        """
        Load a project from the path, `self._project_pathname`.

        Examples:
            >>> load_path = orchid.training_data_path().joinpath('frankNstein_Bakken_UTM13_FEET.ifrac')
            >>> store = ProjectStore(pathname_to_str(load_path))
            >>> store.load_project()
            >>> loaded_project = store.native_project()
            >>> loaded_project.Name
            'frankNstein_Bakken_UTM13_FEET'
        """
        with sac.ScriptAdapterContext():
            reader = ScriptAdapter.CreateProjectFileReader(dot_net.app_settings_path())
            self._native_project = reader.Read(pathname_to_str(self._project_pathname))

    def save_project(self, project):
        """
        Save the specified project to `self._project_pathname`.


        Args:
            project: The project to be saved.

        Examples:
            >>> # Test saving changed project
            >>> load_path = orchid.training_data_path().joinpath('frankNstein_Bakken_UTM13_FEET.ifrac')
            >>> # Use `orchid.core.load_project` to avoid circular dependency with `orchid.project`
            >>> changed_project = orchid.load_project(pathname_to_str(load_path))
            >>> # TODO: move this code to the property eventually, I think.
            >>> with (dnd.disposable(changed_project.dom_object.ToMutable())) as mnp:
            ...     mnp.Name = 'nomen mutatum'
            >>> save_path = load_path.with_name(f'nomen mutatum{load_path.suffix}')
            >>> save_store = ProjectStore(pathname_to_str(save_path))
            >>> save_store.save_project(changed_project)
            >>> save_path.exists()
            True
            >>> with zipfile.ZipFile(save_path) as archive:
            ...     content = json.loads(archive.read('project.json'))
            ...     content['Object']['Name']
            'nomen mutatum'
            >>> save_path.unlink()
            >>> # Test side_effect of `save_project`: `native_project` returns project that was saved
            >>> # I do not expect end users to utilize this side-effect.
            >>> # TODO: Because this code tests a side-effect, an actual unit test might be better.
            >>> load_path = orchid.training_data_path().joinpath('frankNstein_Bakken_UTM13_FEET.ifrac')
            >>> # Use `orchid.core.load_project` to avoid circular dependency with `orchid.project`
            >>> changed_project = orchid.load_project(pathname_to_str(load_path))
            >>> # TODO: move this code to the property eventually, I think.
            >>> with (dnd.disposable(changed_project.dom_object.ToMutable())) as mnp:
            ...     mnp.Name = 'mutatio project'
            >>> save_path = load_path.with_name(f'mutatio project{load_path.suffix}')
            >>> save_store = ProjectStore(pathname_to_str(save_path))
            >>> save_store.save_project(changed_project)
            >>> changed_project.dom_object == save_store.native_project()
            True
            >>> save_path.unlink()
        """
        with sac.ScriptAdapterContext():
            writer = ScriptAdapter.CreateProjectFileWriter()
            use_binary_format = False
            writer.Write(project.dom_object, pathname_to_str(self._project_pathname), use_binary_format)
        self._native_project = project.dom_object

    def optimized_but_possibly_unsafe_save(self, project, maybe_to_pathname: option.Option[Union[str, pathlib.Path]]):
        """
        Saves `project` `to_pathname` is an optimized, but possibly "unsafe" manner.

        This method is unsafe because it only writes some data from `project`; the remainder of the data is simply
        (bulk) copied from the `project_pathname` supplied to the constructor.

        This method assumes that `project` was originally loaded from `project_pathname` and was then changed in
        such a way that the "bulk" data **was not** changed. If this assumption is not true, the project saved in
        `to_pathname` will **not** contain all the changes to `project`.

        Specifically, this method **does not** save changes to data like:

        - Trajectories
        - Treatment curves
        - Monitor curves

        We believe that this method will generally finish more quickly than `save_project`; however, we cannot
        guarantee this behavior. We encourage the developer calling this method to perform her own performance tests
        and to understand if his use case meets the assumptions made by this method.

        Args:
            project: The project to be saved.
            maybe_to_pathname: The "target" pathname for the newly saved data.

        Examples:
            >>> # Test optimized saving of changed project
            >>> load_path = orchid.training_data_path().joinpath('Project_frankNstein_Permian_UTM13_FEET.ifrac')
            >>> # Use `orchid.core.load_project` to avoid circular dependency with `orchid.project`
            >>> changed_project = orchid.load_project(pathname_to_str(load_path))
            >>> # TODO: eventually move this code to a project property, I think.
            >>> with (dnd.disposable(changed_project.dom_object.ToMutable())) as mnp:
            ...     mnp.Name = 'permanet melius'
            >>> save_path = load_path.with_name(f'permanet melius{load_path.suffix}')
            >>> # Remember original path used to load `changed_project`
            >>> changed_project_store = ProjectStore(pathname_to_str(load_path))
            >>> changed_project_store.optimized_but_possibly_unsafe_save(changed_project, option.maybe(save_path))
            >>> save_path.exists()
            True
            >>> with zipfile.ZipFile(save_path) as archive:
            ...     content = json.loads(archive.read('project.json'))
            ...     content['Object']['Name']
            'permanet melius'
            >>> save_path.unlink()
            >>> # Test side_effect of `save_project`: `native_project` returns project that was saved
            >>> # I do not expect end users to utilize this side-effect.
            >>> # TODO: Because this code tests a side-effect, an actual unit test might be better.
            >>> load_path = orchid.training_data_path().joinpath('Project_frankNstein_Permian_UTM13_FEET.ifrac')
            >>> # Use `orchid.core.load_project` to avoid circular dependency with `orchid.project`
            >>> changed_project = orchid.load_project(pathname_to_str(load_path))
            >>> # TODO: move this code to the property eventually, I think.
            >>> with (dnd.disposable(changed_project.dom_object.ToMutable())) as mnp:
            ...     mnp.Name = 'mutatio project melius'
            >>> save_path = load_path.with_name(f'mutatio project melius{load_path.suffix}')
            >>> # Remember original path used to load `changed_project`
            >>> changed_project_store = ProjectStore(pathname_to_str(load_path))
            >>> changed_project_store.optimized_but_possibly_unsafe_save(changed_project, option.maybe(save_path))
            >>> changed_project.dom_object == changed_project_store.native_project()
            True
            >>> save_path.unlink()
            >>> # Test supplying **no** `maybe_to_pathname`
            >>> source_path = orchid.training_data_path().joinpath('Project_frankNstein_Permian_UTM13_FEET.ifrac')
            >>> load_path = source_path.with_name(f'idem filum{source_path.suffix}')
            >>> # ignore returned result for doctest
            >>> _to_path = shutil.copyfile(pathname_to_str(source_path), pathname_to_str(load_path))
            >>> # Use `orchid.core.load_project` to avoid circular dependency with `orchid.project`
            >>> changed_project = orchid.load_project(pathname_to_str(load_path))
            >>> # TODO: eventually move this code to a project property, I think.
            >>> with (dnd.disposable(changed_project.dom_object.ToMutable())) as mnp:
            ...     mnp.Name = 'idem filum'
            >>> changed_project_store = ProjectStore(pathname_to_str(load_path))
            >>> changed_project_store.optimized_but_possibly_unsafe_save(changed_project, option.NONE)
            >>> load_path.exists()
            True
            >>> load_path.unlink()
        """
        with sac.ScriptAdapterContext():
            writer = ScriptAdapter.CreateProjectFileWriter()
            use_binary_format = False
            writer.Write(project.dom_object, pathname_to_str(self._project_pathname),
                         maybe_to_pathname.map_or(pathname_to_str, pathname_to_str(self._project_pathname)),
                         use_binary_format)
        self._native_project = project.dom_object


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#  Copyright 2017-2022 Reveal Energy Services, Inc 
#
#  Licensed under the Apache License, Version 2.0 (the "License"); 
#  you may not use this file except in compliance with the License. 
#  You may obtain a copy of the License at 
#
#      http://www.apache.org/licenses/LICENSE-2.0 
#
#  Unless required by applicable law or agreed to in writing, software 
#  distributed under the License is distributed on an "AS IS" BASIS, 
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
#  See the License for the specific language governing permissions and 
#  limitations under the License. 
#
# This file is part of Orchid and related technologies.
#

"""
Tutorial demonstrating how to add stages to a well using the high-level (Pythonic) API.
"""

import argparse
import logging
import pathlib


import pendulum


import orchid
from orchid import (
    native_stage_adapter as nsa,
)


def add_stages(project, verbosity):
    """
    Add stages to `project`.

    Args:
        project: The `Project` instance to which I add the stages.
        verbosity: The verbosity of the output.
    """
    # Find well to which to add stages
    candidate_well_name = 'Demo_4H'
    candidate_wells = list(project.wells().find_by_display_name(candidate_well_name))
    assert len(candidate_wells) == 1, (f'Expected single well named "{candidate_well_name}".'
                                       f' Found {len(candidate_wells)}.')
    target_well = candidate_wells[0]

    # Create an iterable of stage DTOs (data transfer objects) to append
    # WARNING: The details of these created stages are reasonable but not necessarily consistent with other stage
    # data in the Bakken project.
    # noinspection PyTypeChecker
    stages_to_append = [
        nsa.CreateStageDto(
            36,  # hard-coded to be one greater than largest `order_of_completion_on_well`
            nsa.ConnectionType.PLUG_AND_PERF,
            orchid.make_measurement(orchid.unit_system.UsOilfield.LENGTH, 10966.0),
            orchid.make_measurement(orchid.unit_system.UsOilfield.LENGTH, 11211.0),
            # `pendulum` uses UTC by default for timezone (and UTC required)
            maybe_time_range=pendulum.parse('2018-06-28T22:24:57/2018-06-29T00:10:24'),
            maybe_isip=3420.32 * orchid.unit_registry.psi,
        ),
        nsa.CreateStageDto(
            37,
            nsa.ConnectionType.PLUG_AND_PERF,
            orchid.make_measurement(orchid.unit_system.UsOilfield.LENGTH, 10672.0),
            orchid.make_measurement(orchid.unit_system.UsOilfield.LENGTH, 10917.0),
            maybe_shmin=orchid.make_measurement(orchid.unit_system.UsOilfield.PRESSURE, 2.322),
            # `pendulum` uses UTC by default for timezone (and UTC required)
            maybe_time_range=pendulum.parse('2018-06-29T02:27:56/2018-06-29T04:50:17'),
            maybe_isip=2712.70 * orchid.unit_registry.psi,
        ),
        nsa.CreateStageDto(
            38,
            nsa.ConnectionType.PLUG_AND_PERF,
            orchid.make_measurement(orchid.unit_system.UsOilfield.LENGTH, 10378.0),
            orchid.make_measurement(orchid.unit_system.UsOilfield.LENGTH, 10623.0),
            cluster_count=7,
            # `pendulum` uses UTC by default for timezone (and UTC required)
            maybe_time_range=pendulum.parse('2018-05-29T06:47:15/2018-06-29T08:44:54'),
            maybe_isip=3192.69 * orchid.unit_registry.psi
        ),
    ]

    # Optionally log details of stages to create
    if verbosity >= 1:
        for stage_dto in stages_to_append:
            logging.info('Adding stage')
            logging.info(stage_dto)

    target_well.add_stages(stages_to_append)


def main(cli_args):
    """
    Add stages to an existing project and save changes back to disk.

    Args:
        cli_args: The command line arguments from `argparse.ArgumentParser`.
    """
    logging.basicConfig(level=logging.INFO)

    # Read Orchid project
    project = orchid.load_project(cli_args.input_project)

    add_stages(project, cli_args.verbosity)

    # Save project changes to specified .ifrac file
    orchid.optimized_but_possibly_unsafe_save(project, cli_args.input_project, cli_args.output_project)
    if cli_args.verbosity >= 1:
        logging.info(f'Wrote changes to "{cli_args.output_project}"')


def make_project_path_name(project_dir_name, project_file_name):
    """
    Make a path name to a project.

    Args:
        project_dir_name: The directory name of the project.
        project_file_name: The file name of the project.

    Returns:
        The path name to the .ifrac file for this project.
    """
    return str(project_dir_name.joinpath(project_file_name))


def make_target_file_name_from_source(source_file_name):
    """
    Make a file name for the changed project file name from the original project file name.

    Args:
        source_file_name: The file name of the project originally read.

    Returns:
        The project file name with a `.993` suffix inserted before the `.ifrac` suffix.
    """
    return ''.join([source_file_name.stem, '.993', source_file_name.suffix])


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Adds newly created stages to a well in a project")
    parser.add_argument('-v', '--verbosity', type=int, choices=[0, 1, 2], default=0,
                        help='Increase output verbosity. (Default: 0; that is, least output.)')

    parser.add_argument('input_project', help=f'Path name of project to read.')

    # Although input file must be 'frankNstein_Bakken_UTM13_FEET.v11.ifrac', I use the original version as the
    # default for the correct output file name.
    default_file_name_to_read = pathlib.Path('frankNstein_Bakken_UTM13_FEET.ifrac')
    default_project_path_name_to_read = make_project_path_name(orchid.training_data_path(),
                                                               default_file_name_to_read)
    default_file_name_to_write = make_target_file_name_from_source(default_file_name_to_read)
    default_project_path_name_to_write = make_project_path_name(orchid.training_data_path(),
                                                                default_file_name_to_write)
    parser.add_argument('-o', '--output_project', default=default_project_path_name_to_write,
                        help=f'Filename of project to write. (Default: {default_project_path_name_to_write}')

    args = parser.parse_args()
    main(args)



# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# This file is part of Orchid and related technologies.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#


from abc import abstractmethod
from collections import namedtuple
from enum import Enum
import numbers

import toolz.curried as toolz

from orchid import (
    measurement as om,
    physical_quantity as opq,
)


# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics import UnitSystem as NetUnitSystem
# noinspection PyUnresolvedReferences
import UnitsNet


# I use this private class to distinguish units that measure **different** physical quantities but use the
# same measurement unit. Currently, this applies to DENSITY and PRESSURE.
_AboutUnit = namedtuple('AboutUnit', ['unit', 'physical_quantity'])


class UnitSystem(Enum):

    def __str__(self):
        """
        Return a string representation of the member.

        Returns:
            The string representing the enumeration member.
        """
        return f'{self.value.unit}'

    def abbreviation(self):
        """
        Return the abbreviation for a unit.

        Returns:
            The abbreviation of 'unit'.
        """
        return f'{self.value.unit:~P}'

    @abstractmethod
    def system_name(self):
        raise NotImplementedError()


class Common(UnitSystem):
    """The enumeration of units common to both U. S. oilfield and metric unit systems."""

    ANGLE = _AboutUnit(om.registry.deg, opq.PhysicalQuantity.ANGLE)
    DURATION = _AboutUnit(om.registry.min, opq.PhysicalQuantity.DURATION)

    def system_name(self):
        return 'Common'


class UsOilfield(UnitSystem):
    """The enumeration of U. S. oilfield units available via the Orchid Python API."""

    DENSITY = _AboutUnit(om.registry.pound / om.registry.ft ** 3, opq.PhysicalQuantity.DENSITY)
    ENERGY = _AboutUnit(om.registry.foot_pound, opq.PhysicalQuantity.ENERGY)
    FORCE = _AboutUnit(om.registry.pound_force, opq.PhysicalQuantity.FORCE)
    LENGTH = _AboutUnit(om.registry.foot, opq.PhysicalQuantity.LENGTH)
    MASS = _AboutUnit(om.registry.pound, opq.PhysicalQuantity.MASS)
    POWER = _AboutUnit(om.registry.horsepower, opq.PhysicalQuantity.POWER)
    PRESSURE = _AboutUnit(om.registry.pound_force_per_square_inch, opq.PhysicalQuantity.PRESSURE)
    PROPPANT_CONCENTRATION = _AboutUnit(om.registry.pound / om.registry.gallon,
                                        opq.PhysicalQuantity.PROPPANT_CONCENTRATION)
    SLURRY_RATE = _AboutUnit(om.registry.oil_barrel / om.registry.minute, opq.PhysicalQuantity.SLURRY_RATE)
    TEMPERATURE = _AboutUnit(om.registry.degree_Fahrenheit, opq.PhysicalQuantity.TEMPERATURE)
    VOLUME = _AboutUnit(om.registry.oil_barrel, opq.PhysicalQuantity.VOLUME)

    def system_name(self):
        return 'USOilfield'


class Metric(UnitSystem):
    """The enumeration of metric units available via the Orchid Python API."""

    DENSITY = _AboutUnit(om.registry.kilogram / om.registry.meter ** 3, opq.PhysicalQuantity.DENSITY)
    ENERGY = _AboutUnit(om.registry.joule, opq.PhysicalQuantity.ENERGY)
    FORCE = _AboutUnit(om.registry.newton, opq.PhysicalQuantity.FORCE)
    LENGTH = _AboutUnit(om.registry.meter, opq.PhysicalQuantity.LENGTH)
    MASS = _AboutUnit(om.registry.kilogram, opq.PhysicalQuantity.MASS)
    POWER = _AboutUnit(om.registry.watt, opq.PhysicalQuantity.POWER)
    PRESSURE = _AboutUnit(om.registry.kilopascal, opq.PhysicalQuantity.PRESSURE)
    PROPPANT_CONCENTRATION = _AboutUnit(om.registry.kilogram / om.registry.meter ** 3,
                                        opq.PhysicalQuantity.PROPPANT_CONCENTRATION)
    SLURRY_RATE = _AboutUnit(om.registry.meter ** 3 / om.registry.minute, opq.PhysicalQuantity.SLURRY_RATE)
    TEMPERATURE = _AboutUnit(om.registry.degree_Celsius, opq.PhysicalQuantity.TEMPERATURE)
    VOLUME = _AboutUnit((om.registry.m ** 3), opq.PhysicalQuantity.VOLUME)

    def system_name(self):
        return 'Metric'


def as_unit_system(net_unit_system: UnitSystem):
    if net_unit_system == NetUnitSystem.USOilfield():
        return UsOilfield
    elif net_unit_system == NetUnitSystem.Metric():
        return Metric
    else:
        raise ValueError(f'Unrecognized unit system: {net_unit_system}')


@toolz.curry
def make_measurement(unit: UnitSystem, magnitude: numbers.Real) -> om.Quantity:
    """
    Construct a measurement.

    This function provides a "functional" mechanism to create Measurement instances. It is more common to
    create a sequence of measurements from a sequence of numbers and a **single** unit. By putting the `unit`
    argument first in the function arguments, it allows callers to write code similar to the following:

    > make_length_measurement = make_measurement(units.UsOilfield.LENGTH)
    > length_measurements = [make_length_measurement(l) for l in lengths]
    > # or toolz.map(make_length_measurement, lengths)

    Args:
        unit: The unit of this measurement.
        magnitude: The magnitude of the measurement.

    Returns:
        The created `pint` `Quantity` instance.
    """
    return om.Quantity(magnitude, unit.value.unit)


make_us_oilfield_length_measurement = make_measurement(UsOilfield.LENGTH)
make_us_oilfield_pressure_measurement = make_measurement(UsOilfield.PRESSURE)

make_metric_length_measurement = make_measurement(Metric.LENGTH)
make_metric_pressure_measurement = make_measurement(Metric.PRESSURE)


def is_length_unit(candidate) -> bool:
    return candidate.check('[length]')


def is_pressure_unit(candidate) -> bool:
    return candidate.check('[pressure]')


def abbreviation(unit: UnitSystem):
    """
    Return the abbreviation of `unit`.

    This function provides a functional interface for calculating abbreviations of `UnitSystem` members.

    Args:
        unit: The `UnitSystem` member whose abbreviation is sought.

    Returns:
        The abbreviation for `unit`.
    """
    return unit.abbreviation()



#  Copyright (c) 2017-2022 Reveal Energy Services, Inc 
#
#  Licensed under the Apache License, Version 2.0 (the "License"); 
#  you may not use this file except in compliance with the License. 
#  You may obtain a copy of the License at 
#
#      http://www.apache.org/licenses/LICENSE-2.0 
#
#  Unless required by applicable law or agreed to in writing, software 
#  distributed under the License is distributed on an "AS IS" BASIS, 
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
#  See the License for the specific language governing permissions and 
#  limitations under the License. 
#
# This file is part of Orchid and related technologies.
#

from .dot_net import prepare_imports
prepare_imports()

# High-level API
from .core import load_project, save_project, optimized_but_possibly_unsafe_save

# Helpful constants
from .native_treatment_curve_adapter import TreatmentCurveTypes
from .native_time_series_adapter import TimeSeriesCurveTypes
from .net_date_time import UTC

# Helpful functions
from .convert import to_unit
from .measurement import registry as unit_registry
from .native_treatment_calculations import (median_treating_pressure, pumped_fluid_volume, total_proppant_mass)
from .reference_origins import WellReferenceFrameXy
from .unit_system import abbreviation, make_measurement

# Only for training data
from .configuration import training_data_path
